from __future__ import annotations

import io
import sys
import typing as t
import logging
from pathlib import Path

import loguru

from starwhale.utils.fs import ensure_dir
from starwhale.utils.log import StreamWrapper

from .base import _TrackSource


def hook_frameworks() -> None:
    ...


def hook_python_libs() -> None:
    ...


class ConsoleLogHook:
    def __init__(self, log_dir: Path, track_id: str, rotation: str = "100 MB") -> None:
        self.log_dir = log_dir

        self._stdout_changed: bool = False
        self._stderr_changed: bool = False
        self._original_stdout: t.Optional[io.TextIOWrapper] = None
        self._original_stderr: t.Optional[io.TextIOWrapper] = None

        self._rotation = rotation
        self._track_id = track_id
        self._logger: t.Optional[loguru.Logger] = None
        self._log_handler_id: t.Optional[int] = None

    def _setup_logger(self) -> loguru.Logger:
        ensure_dir(self.log_dir)
        loguru.logger.remove()
        self._log_handler_id = loguru.logger.add(
            self.log_dir / "console-{time}.log",
            rotation=self._rotation,
            backtrace=True,
            diagnose=True,
            serialize=True,
        )
        _logger = loguru.logger.bind(
            source=_TrackSource.USER.value,
            track_id=self._track_id,
        )
        return _logger

    def install(self) -> None:
        if self._logger is None:
            self._logger = self._setup_logger()

        if not isinstance(sys.stdout, StreamWrapper) and isinstance(
            sys.stdout, io.TextIOWrapper
        ):
            self._original_stdout = sys.stdout
            sys.stdout = StreamWrapper(sys.stdout, self._logger, logging.INFO)  # type: ignore
            self._stdout_changed = True

        if not isinstance(sys.stderr, StreamWrapper) and isinstance(
            sys.stderr, io.TextIOWrapper
        ):
            self._original_stderr = sys.stderr
            sys.stderr = StreamWrapper(sys.stderr, self._logger, logging.WARN)  # type: ignore
            self._stderr_changed = True

        # TODO: redirect python standard logger

    def uninstall(self) -> None:
        if self._log_handler_id is not None and self._logger is not None:
            self._logger.remove(self._log_handler_id)
            self._logger = None
            self._log_handler_id = None

        if self._stdout_changed and self._original_stdout is not None:
            sys.stdout = self._original_stdout
            self._stdout_changed = False
            self._original_stdout = None

        if self._stderr_changed and self._original_stderr is not None:
            sys.stderr = self._original_stderr
            self._stderr_changed = False
            self._original_stderr = None



from starwhale.core.dataset.type import (
    Line,
    Link,
    Text,
    Audio,
    Image,
    Point,
    Video,
    Binary,
    Polygon,
    LinkAuth,
    LinkType,
    MIMEType,
    ClassLabel,
    S3LinkAuth,
    BoundingBox,
    NumpyBinary,
    BoundingBox3D,
    GrayscaleImage,
    LocalFSLinkAuth,
    DefaultS3LinkAuth,
    COCOObjectAnnotation,
)

from .model import Dataset

__all__ = [
    "Link",
    "LinkAuth",
    "DefaultS3LinkAuth",
    "LocalFSLinkAuth",
    "S3LinkAuth",
    "MIMEType",
    "LinkType",
    "Binary",
    "NumpyBinary",
    "Text",
    "Line",
    "Point",
    "Polygon",
    "Audio",
    "Video",
    "Image",
    "ClassLabel",
    "BoundingBox",
    "BoundingBox3D",
    "GrayscaleImage",
    "COCOObjectAnnotation",
    "Dataset",
]



import os
import sys
import logging

from rich import traceback
from loguru import logger

from starwhale.consts import ENV_LOG_LEVEL, ENV_LOG_VERBOSE_COUNT


def init_logger(verbose: int) -> None:
    fmt = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <level>{message}</level>"
    if verbose <= 0:
        lvl = logging.WARNING
    elif verbose == 1:
        lvl = logging.INFO
    else:
        lvl = logging.DEBUG

    lvl_name = logging.getLevelName(lvl)
    os.environ[ENV_LOG_LEVEL] = lvl_name
    os.environ[ENV_LOG_VERBOSE_COUNT] = str(verbose)

    # TODO: custom debug for tb install
    traceback.install(show_locals=True, max_frames=1, width=200)

    logger.remove()
    logger.add(
        sys.stderr,
        level=lvl_name,
        colorize=True,
        backtrace=True,
        diagnose=True,
        catch=True,
        format=fmt,
    )
    logger.debug(f"verbosity: {verbose}, log level: {lvl_name}")



from __future__ import annotations

import sys
import typing as t
import inspect
import numbers
import threading
from copy import deepcopy
from pathlib import Path
from functools import wraps
from collections import defaultdict

import yaml
from loguru import logger

from starwhale.consts import DEFAULT_JOB_NAME, DecoratorInjectAttr
from starwhale.core.job import dag
from starwhale.utils.fs import ensure_file
from starwhale.utils.load import load_module
from starwhale.utils.error import NoSupportError
from starwhale.core.job.step import Step
from starwhale.core.job.context import Context

AFTER_LOAD_HOOKS: t.Dict[str, t.Callable] = defaultdict()
_T_JOBS = t.Dict[str, t.List[Step]]
_jobs_global: _T_JOBS = defaultdict(list)
_jobs_global_lock = threading.Lock()


# TODO: refactor _do_resource_transform, move it into Step
def _do_resource_transform(resources: t.Optional[t.Dict[str, t.Any]]) -> t.List[t.Dict]:
    attribute_names = ["request", "limit"]
    resource_names: t.Dict[str, t.List] = {
        "cpu": [int, float],
        "nvidia.com/gpu": [int],
        "memory": [int, float],
    }

    resources = resources or {}
    results = []
    for _name, _resource in resources.items():
        if _name not in resource_names:
            raise RuntimeError(
                f"resources name is illegal, name must in {resource_names.keys()}"
            )

        if isinstance(_resource, numbers.Number):
            resources[_name] = {"request": _resource, "limit": _resource}
        elif isinstance(_resource, dict):
            if not all(n in attribute_names for n in _resource):
                raise RuntimeError(
                    f"resources value is illegal, attribute's name must in {attribute_names}"
                )
        else:
            raise RuntimeError(
                "resources value is illegal, attribute's type must be number or dict"
            )

        for _k, _v in resources[_name].items():
            if type(_v) not in resource_names[_name]:
                raise RuntimeError(
                    f"resource:{_name} only support type:{resource_names[_name]}, but now is {type(_v)}"
                )
            if _v <= 0:
                raise RuntimeError(
                    f"{_k} only supports non-negative number, but now is {_v}"
                )
        results.append(
            {
                "type": _name,
                "request": resources[_name]["request"],
                "limit": resources[_name]["limit"],
            }
        )
    return results


def step(
    job_name: str = DEFAULT_JOB_NAME,
    resources: t.Optional[t.Dict[str, t.Any]] = None,
    concurrency: int = 1,
    task_num: int = 1,
    needs: t.Optional[t.List[str]] = None,
    extra_args: t.Optional[t.List] = None,
    extra_kwargs: t.Optional[t.Dict] = None,
    name: str = "",
) -> t.Callable:
    def decorator(func: t.Callable) -> t.Callable:
        module = inspect.getmodule(func)
        if module is None:
            raise RuntimeError(f"cannot get module name from func: {func}")

        if inspect.isclass(func):
            raise NoSupportError(f"step decorator no support class: {func}")

        cls_name, _, func_name = func.__qualname__.rpartition(".")
        if "." in cls_name:
            raise NoSupportError(
                f"step decorator no supports inner class method:{func.__qualname__}"
            )

        _step = Step(
            job_name=job_name,
            name=name or func_name,
            func_name=func_name,
            cls_name=cls_name,
            module_name=module.__name__,
            concurrency=concurrency,
            task_num=task_num,
            needs=needs,
            resources=_do_resource_transform(resources),
            extra_args=extra_args,
            extra_kwargs=extra_kwargs,
        )

        global _jobs_global, _jobs_global_lock
        with _jobs_global_lock:
            _jobs_global[job_name].append(_step)

        setattr(func, DecoratorInjectAttr.Step, True)
        return func

    return decorator


def _validate_jobs_dag(jobs: _T_JOBS) -> None:
    for steps in jobs.values():
        _vertices: t.List[str] = []
        _edges: t.Dict[str, str] = {}
        for _step in steps:
            _vertices.append(_step.name)
            for _pre in _step.needs:
                if _pre:
                    _edges[_pre] = _step.name

        dag.generate_dag(_vertices, _edges)


def _preload_to_register_jobs(run_handler: str, workdir: Path) -> None:
    """
    run_handler formats:
    - to.one.module
    - to.one.module:function  --> with @step, @predict decorator
    - to.one.module:ArbitraryClass  --> use @step or @predict to decorate some class methods
    - to.one.module:PipeHandlerSubClass
    """
    module_name, _, func_or_cls_name = run_handler.partition(":")

    # reload for the multi model.build in one python process
    if module_name in sys.modules:
        del sys.modules[module_name]

    module = load_module(module_name, workdir)
    func_or_cls = getattr(module, func_or_cls_name, None)

    if func_or_cls is None:
        logger.debug(f"preload module-{module_name}")
        return

    for _, _h in AFTER_LOAD_HOOKS.items():
        _h(run_handler, func_or_cls)


def generate_jobs_yaml(
    run_handler: str, workdir: t.Union[Path, str], yaml_path: t.Union[Path, str]
) -> None:
    workdir = Path(workdir)
    logger.debug(f"ingest steps from run_handler {run_handler} at {workdir}")

    _preload_to_register_jobs(run_handler, workdir)

    global _jobs_global, _jobs_global_lock
    with _jobs_global_lock:
        jobs = deepcopy(_jobs_global)

        _names = list(_jobs_global.keys())
        [_jobs_global.__delitem__(n) for n in _names]  # type: ignore[func-returns-value]

    if not jobs:
        raise RuntimeError("not found any jobs")

    _validate_jobs_dag(jobs)
    ensure_file(
        yaml_path,
        yaml.safe_dump(
            {job_name: [s.asdict() for s in steps] for job_name, steps in jobs.items()},
            default_flow_style=False,
        ),
        parents=True,
    )


def pass_context(func: t.Any) -> t.Any:
    @wraps(func)
    def wrap_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
        kwargs["context"] = Context.get_runtime_context()
        return func(*args, **kwargs)

    return wrap_func



import sys
import typing as t

import click

from starwhale.utils.cli import AliasedGroup

from .view import get_term_view, InstanceTermView


@click.group(
    "instance",
    cls=AliasedGroup,
    help="Starwhale Instance management, login and select standalone or cloud instance",
)
@click.pass_context
def instance_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@instance_cmd.command("select", aliases=["use"])
@click.argument("instance")
def _select(instance: str) -> None:
    """
    Select current default instance

    INSTANCE: use alias name to select
    """
    # TODO: support URI select
    InstanceTermView().select(instance)


@instance_cmd.command("login")
@click.argument("instance", default="")
@click.option("--username")
@click.option("--password")
@click.option("--token", help="Login token")
@click.option("--alias", type=str, help="Starwhale instance alias name", required=True)
def _login(instance: str, username: str, password: str, token: str, alias: str) -> None:
    """Login Starwhale Instance

    * INSTANCE: Instance URI, if ignore it, swcli will login current selected instance.
    """
    if not bool(password and username) ^ bool(token):
        click.echo("token or password+username, only choose one type")
        sys.exit(1)

    if token:
        kw = {"token": token}
    else:
        kw = {"username": username, "password": password}

    InstanceTermView().login(instance, alias, **kw)


@instance_cmd.command("logout")
@click.argument("instance", default="")
def _logout(instance: str) -> None:
    """
    Logout Starwhale instance

    * INSTANCE: instance alias name or uri, if ignore it, swcli will logout current selected instance.
    """
    click.confirm("Do you want to continue?", abort=True)
    InstanceTermView().logout(instance)


@instance_cmd.command("list", aliases=["ls"], help="List login Starwhale instances")
@click.pass_obj
def _list(view: t.Type[InstanceTermView]) -> None:
    view().list()


@instance_cmd.command("info", help="Show instance details")
@click.argument("instance", default="")
@click.pass_obj
def _info(view: t.Type[InstanceTermView], instance: str) -> None:
    view().info(instance)



import os
import re
import urllib
import threading
from enum import Enum, unique
from typing import Any, Dict, List, Union, Callable, Iterator, Optional
from functools import lru_cache

import dill
import requests
from loguru import logger

from starwhale.consts import VERSION_PREFIX_CNT, STANDALONE_INSTANCE
from starwhale.consts.env import SWEnv
from starwhale.utils.retry import http_retry
from starwhale.utils.config import SWCliConfigMixed

from . import data_store


class Logger:
    def _init_writers(self, tables: List[str]) -> None:
        self._writers: Dict[str, Optional[data_store.TableWriter]] = {
            table: None for table in tables
        }
        self._lock = threading.Lock()

    def close(self) -> None:
        with self._lock:
            exceptions: List[Exception] = []
            for writer in self._writers.values():
                if writer is None:
                    continue

                try:
                    writer.close()
                except Exception as e:
                    logger.exception(f"{writer} exception: {e}")
                    exceptions.append(e)

            if exceptions:
                raise Exception(*exceptions)

    def _fetch_writer(self, table_name: str) -> data_store.TableWriter:
        with self._lock:
            if table_name not in self._writers:
                self._writers.setdefault(table_name, None)
            writer = self._writers[table_name]
            if writer is None:
                _store = getattr(self, "_data_store", None)
                writer = data_store.TableWriter(table_name, data_store=_store)
                self._writers[table_name] = writer
        return writer

    def _log(self, table_name: str, record: Dict[str, Any]) -> None:
        writer = self._fetch_writer(table_name)
        writer.insert(record)

    def _flush(self, table_name: str) -> str:
        with self._lock:
            writer = self._writers.get(table_name)
            if writer is None:
                return ""
        return writer.flush()

    def _delete(self, table_name: str, key: Any) -> None:
        writer = self._fetch_writer(table_name)
        writer.delete(key)


def _serialize(data: Any) -> Any:
    return dill.dumps(data)


def _deserialize(data: bytes) -> Any:
    return dill.loads(data)


table_name_formatter: Callable[
    [Union[str, int], str], str
] = lambda project, table: f"project/{project}/{table}"


def _gen_storage_table_name(
    project: Union[str, int], table: str, instance_uri: str = ""
) -> str:
    _instance_uri = instance_uri or os.getenv(SWEnv.instance_uri)
    if (
        _instance_uri is None
        or _instance_uri == STANDALONE_INSTANCE
        or isinstance(project, int)
    ):
        return table_name_formatter(project, table)
    else:
        return table_name_formatter(
            _get_remote_project_id(_instance_uri, project), table
        )


@lru_cache(maxsize=None)
@http_retry
def _get_remote_project_id(instance_uri: str, project: str) -> Any:
    resp = requests.get(
        urllib.parse.urljoin(instance_uri, f"/api/v1/project/{project}"),
        headers={
            "Content-Type": "application/json; charset=utf-8",
            "Authorization": (
                SWCliConfigMixed().get_sw_token(instance=instance_uri)
                or os.getenv(SWEnv.instance_token, "")
            ),
        },
        timeout=60,
    )
    resp.raise_for_status()
    return resp.json().get("data", {})["id"]


class Evaluation(Logger):
    def __init__(self, eval_id: str, project: str, instance: str = ""):
        if not eval_id:
            raise RuntimeError("eval id should not be None")
        if re.match(r"^[A-Za-z0-9-_]+$", eval_id) is None:
            raise RuntimeError(
                f"invalid eval id {eval_id}, only letters(A-Z, a-z), digits(0-9), hyphen('-'), and underscore('_') are allowed"
            )
        if not project:
            raise RuntimeError("project is not set")

        self.eval_id = eval_id
        self.project = project
        self.instance = instance
        self._tables: Dict[str, str] = {}
        self._eval_table_name: Callable[
            [str], str
        ] = (
            lambda name: f"eval/{self.eval_id[:VERSION_PREFIX_CNT]}/{self.eval_id}/{name}"
        )
        self._eval_summary_table_name = "eval/summary"
        self._data_store = data_store.get_data_store(instance_uri=instance)
        self._init_writers([])

    def _get_storage_table_name(self, table: str) -> str:
        with self._lock:
            _table_name = self._tables.get(table)
            if _table_name is None:
                _table_name = _gen_storage_table_name(
                    project=self.project,
                    table=table,
                    instance_uri=self.instance,
                )
                self._tables[table] = _table_name
        return _table_name

    def _log(self, table_name: str, record: Dict[str, Any]) -> None:
        _storage_table_name = self._get_storage_table_name(table_name)
        super()._log(_storage_table_name, record)

    def _get(self, table_name: str) -> Iterator[Dict[str, Any]]:
        return self._data_store.scan_tables(
            [data_store.TableDesc(self._get_storage_table_name(table=table_name))]
        )

    def _flush(self, table_name: str) -> str:
        _storage_table_name = self._get_storage_table_name(table_name)
        return super()._flush(_storage_table_name)

    def log_result(
        self,
        data_id: Union[int, str],
        result: Any,
        serialize: bool = False,
        **kwargs: Any,
    ) -> None:
        record = {"id": data_id, "result": _serialize(result) if serialize else result}
        for k, v in kwargs.items():
            record[k.lower()] = _serialize(v) if serialize else v

        self._log(self._eval_table_name("results"), record)

    def log_metrics(
        self, metrics: Optional[Dict[str, Any]] = None, **kwargs: Any
    ) -> None:
        record = {"id": self.eval_id}
        # TODO: without if else?
        if metrics is not None:
            for k, v in metrics.items():
                k = k.lower()
                if k != "id":
                    record[k] = v
        else:
            for k, v in kwargs.items():
                record[k.lower()] = v

        self._log(self._eval_summary_table_name, record)

    def log(self, table_name: str, **kwargs: Any) -> None:
        record = {}
        for k, v in kwargs.items():
            record[k.lower()] = v
        self._log(self._eval_table_name(table_name), record)

    def get_results(self, deserialize: bool = False) -> Iterator[Dict[str, Any]]:
        for data in self._get(self._eval_table_name("results")):
            if deserialize:
                for _k, _v in data.items():
                    if _k == "id":
                        continue
                    data[_k] = _deserialize(_v)
            yield data

    def get_metrics(self) -> Dict[str, Any]:
        for metrics in self._get(self._eval_summary_table_name):
            if metrics["id"] == self.eval_id:
                return metrics

        return {}

    def get(self, table_name: str) -> Iterator[Dict[str, Any]]:
        return self._get(self._eval_table_name(table_name))

    def flush_result(self) -> None:
        self._flush(self._eval_table_name("results"))

    def flush_metrics(self) -> None:
        self._flush(self._eval_summary_table_name)

    def flush(self, table_name: str) -> None:
        self._flush(self._eval_table_name(table_name))


@unique
class DatasetTableKind(Enum):
    META = "meta"
    INFO = "info"


class Dataset(Logger):
    def __init__(
        self,
        dataset_name: str,
        project: str,
        dataset_scan_revision: str = "",
        instance_name: str = "",
        token: str = "",
        kind: DatasetTableKind = DatasetTableKind.META,
    ) -> None:
        if not dataset_name:
            raise RuntimeError("id should not be None")

        if not project:
            raise RuntimeError("project is not set")

        self.dataset_scan_revision = dataset_scan_revision
        self.project = project
        self.kind = kind

        # _current is only holder part of the dataset table name
        self._table_name = _gen_storage_table_name(
            project=project,
            table=f"dataset/{dataset_name}/_current/{kind.value}",
            instance_uri=instance_name,
        )
        self._data_store = data_store.get_data_store(instance_name, token)
        self._init_writers([self._table_name])

    def put(self, data_id: Union[str, int], **kwargs: Any) -> None:
        record = {"id": data_id}
        for k, v in kwargs.items():
            record[k.lower()] = v
        self._log(self._table_name, record)

    def delete(self, data_id: Union[str, int]) -> None:
        self._delete(self._table_name, data_id)

    def scan(
        self,
        start: Any,
        end: Any,
        end_inclusive: bool = False,
        revision: str = "",
    ) -> Iterator[Dict[str, Any]]:
        return self._data_store.scan_tables(
            [
                data_store.TableDesc(
                    table_name=self._table_name,
                    revision=revision or self.dataset_scan_revision,
                )
            ],
            start=start,
            end=end,
            end_inclusive=end_inclusive,
        )

    def scan_id(
        self,
        start: Any,
        end: Any,
        end_inclusive: bool = False,
        revision: str = "",
    ) -> Iterator[Any]:
        return self._data_store.scan_tables(
            [
                data_store.TableDesc(
                    table_name=self._table_name,
                    columns=["id"],
                    revision=revision or self.dataset_scan_revision,
                )
            ],
            start=start,
            end=end,
            end_inclusive=end_inclusive,
        )

    def flush(self) -> str:
        return self._flush(self._table_name)

    def __str__(self) -> str:
        return f"Dataset Wrapper, table:{self._table_name}, store:{self._data_store}, kind: {self.kind}, revision: {self.dataset_scan_revision}"

    __repr__ = __str__



import os
import errno
import shutil
import typing as t
import difflib
import hashlib
import tarfile
from enum import IntEnum
from pathlib import Path

from starwhale.utils import console, timestamp_to_datatimestr
from starwhale.utils.error import FormatError, ExistedError, NotFoundError
from starwhale.utils.process import check_call

BLAKE2B_SIGNATURE_ALGO = "blake2b"
_MIN_GUESS_NAME_LENGTH = 5


class FilePosition(IntEnum):
    START = 0
    END = -1


def ensure_file(
    path: t.Union[str, Path],
    content: t.Union[str, bytes],
    mode: int = 0o644,
    parents: bool = False,
) -> None:
    p = Path(path)
    if parents:
        p.parent.mkdir(parents=True, exist_ok=True)
    bin_mode = isinstance(content, bytes)
    try:
        with p.open("rb" if bin_mode else "r") as f:
            _saved = f.read()
    except IOError as e:
        if e.errno == errno.ENOENT:
            # no such file or directory
            _saved = ""
        else:
            raise

    if _saved != content or not p.exists():
        # TODO: add timestamp for tmp file
        _tmp_f = p.parent / f".{p.name}.tmp"
        if isinstance(content, bytes):
            _tmp_f.write_bytes(content)
        elif isinstance(content, str):
            _tmp_f.write_text(content)
        else:
            raise TypeError(
                f"content({type(content)}-{content}) only accepts bytes or str type"
            )
        # TODO: check whether rename atomic
        _tmp_f.rename(path)

    os.chmod(path, mode)


def empty_dir(
    p: t.Union[str, Path],
    ignore_errors: bool = False,
    onerror: t.Optional[t.Callable] = None,
) -> None:
    if not p:
        return

    path = Path(p)
    if not path.exists() or path.resolve() == Path("/"):
        return

    def _self_empty() -> None:
        if path.is_dir():
            shutil.rmtree(str(path.resolve()), ignore_errors, onerror)
        else:
            path.unlink()

    def _sudo_empty() -> None:
        console.print(
            f":bell: try to use [red bold]SUDO[/] privilege to empty : {path}"
        )
        check_call(f"sudo rm -rf {path}", shell=True)

    try:
        _self_empty()
    except PermissionError:
        _sudo_empty()


def ensure_dir(
    path: t.Union[str, Path], mode: int = 0o755, recursion: bool = True
) -> None:
    p = Path(path)
    if p.exists() and p.is_dir():
        return

    try:
        if recursion:
            os.makedirs(path, mode)
        else:
            os.mkdir(path, mode)
    except OSError as e:
        if e.errno != errno.EEXIST:
            # TODO: add more human friendly exception log
            raise


def ensure_link(src: t.Union[str, Path], dest: t.Union[str, Path]) -> None:
    src, dest = str(src), str(dest)
    dirname = os.path.dirname(dest)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)

    if os.path.lexists(dest):
        if os.path.islink(dest) and os.readlink(dest) == src:
            return
        os.unlink(dest)
    os.symlink(src, dest)


def blake2b_file(fpath: t.Union[str, Path]) -> str:
    _chunk_size = 8192
    fpath = Path(fpath)
    # blake2b is more faster and better than md5,sha1,sha2
    _hash = hashlib.blake2b(digest_size=64)

    with fpath.open("rb") as f:
        _chunk = f.read(_chunk_size)
        while _chunk:
            _hash.update(_chunk)
            _chunk = f.read(_chunk_size)

    return _hash.hexdigest()


def blake2b_content(content: bytes) -> str:
    _hash = hashlib.blake2b(digest_size=64)
    _hash.update(content)
    return _hash.hexdigest()


def cmp_file_content(
    base_path: t.Union[str, Path], cmp_path: t.Union[str, Path]
) -> t.List[str]:
    base_path = Path(base_path)
    cmp_path = Path(cmp_path)
    res = []
    with base_path.open("r") as base, cmp_path.open("r") as cmp:
        diff = difflib.unified_diff(
            base.readlines(),
            cmp.readlines(),
            fromfile=base_path.name,
            tofile=cmp_path.name,
            n=0,
        )
        for line in diff:

            def skip(content: str) -> bool:
                prefixes = ["---", "+++", "@@", " "]
                return not any([content.startswith(prefix) for prefix in prefixes])

            if skip(content=line):
                res.append(line)
    return res


def get_path_created_time(p: Path) -> str:
    created_at = os.path.getctime(str(p.absolute()))
    return timestamp_to_datatimestr(created_at)


def guess_real_path(
    rootdir: Path, name: str, ftype: str = ""
) -> t.Tuple[Path, str, bool]:
    # TODO: support more guess method, such as tag
    _path = rootdir / name
    if _path.exists():
        return _path, name, True

    if not rootdir.exists():
        return _path, name, False

    if len(name) < _MIN_GUESS_NAME_LENGTH:
        return _path, name, False

    ftype = ftype.strip()
    for fd in rootdir.iterdir():
        if ftype and not fd.name.endswith(ftype):
            continue

        if fd.name.startswith(name) or name.startswith(fd.name):
            return fd, fd.name.rsplit(ftype, 1)[0] if ftype else fd.name, True
    else:
        return _path, name, False


def move_dir(src: Path, dest: Path, force: bool = False) -> t.Tuple[bool, str]:
    if not src.exists():
        return False, f"src:{src} not found"

    if dest.exists() and not force:
        return False, f"dest:{dest} existed"

    ensure_dir(dest.parent)

    try:
        shutil.move(str(src.absolute()), str(dest.absolute()))
    except Exception as e:
        return False, f"failed to move {src} -> {dest}, reason: {e}"
    else:
        return True, f"{src} move to {dest}"


def extract_tar(tar_path: Path, dest_dir: Path, force: bool = False) -> None:
    if not tar_path.exists():
        raise NotFoundError(tar_path)

    if dest_dir.exists() and not force:
        raise ExistedError(str(dest_dir))

    empty_dir(dest_dir)
    ensure_dir(dest_dir)

    with tarfile.open(tar_path, "r") as tar:
        for member in tar.getmembers():
            if not is_within_dir(dest_dir, dest_dir / member.name):
                raise Exception("Attempted path traversal in tar file")
        tar.extractall(path=str(dest_dir.absolute()))


def copy_file(src: Path, dest: Path) -> None:
    if not src.exists():
        raise NotFoundError(src)

    if not src.is_file():
        raise FormatError(f"{src} is not file type")

    ensure_dir(dest.parent)
    shutil.copyfile(str(src.absolute()), str(dest.absolute()))


def is_within_dir(parent: t.Union[str, Path], child: t.Union[str, Path]) -> bool:
    parent = str(parent)
    child = str(child)

    abs_parent = os.path.abspath(parent)
    abs_child = os.path.abspath(child)

    prefix = os.path.commonprefix([abs_parent, abs_child])
    return prefix == abs_parent


def file_stat(path: t.Union[str, Path]) -> os.stat_result:
    path = str(path)
    abs_path = os.path.abspath(path)
    return Path(abs_path).stat()



import typing as t
import collections

import numpy
import tensorflow as tf

import starwhale.core.dataset.type as sw_type
from starwhale.utils.error import NoSupportError
from starwhale.api._impl.dataset import Dataset

__all__ = ["to_tf_dataset"]

# https://www.tensorflow.org/api_docs/python/tf/dtypes
_TYPE_DICT: t.Dict[t.Any, tf.DType] = {
    int: tf.dtypes.int64,
    float: tf.dtypes.float64,
    complex: tf.dtypes.complex128,
    bool: tf.dtypes.bool,
    str: tf.dtypes.string,
    bytes: tf.dtypes.string,
}

# TODO: sw_type supports Sequence Type to store list[dict] type: [{"a": 1}, {"a":2}]


def _transform(data: t.Any) -> t.Any:
    dtype = type(data)
    if dtype in _TYPE_DICT:
        return data

    if isinstance(data, (sw_type.BaseArtifact, sw_type.BoundingBox)):
        return data.to_numpy()
    elif isinstance(data, sw_type.Text):
        return data.to_str()
    elif isinstance(data, collections.abc.Mapping):  # type: ignore
        try:
            return dtype({k: _transform(v) for k, v in data.items()})
        except TypeError:
            # The mapping type may not support __init__(iterable)
            return {k: _transform(v) for k, v in data.items()}
    elif isinstance(data, collections.abc.Sequence):  # type: ignore
        try:
            return dtype([_transform(d) for d in data])
        except TypeError:
            # The sequence type may not support __init__(iterable), (e.g.: range)
            return [_transform(d) for d in data]
    else:
        return data


def _inspect_spec(data: t.Any) -> t.Union[tf.TensorSpec, t.Dict]:
    dtype = type(data)
    if dtype in _TYPE_DICT:
        return tf.TensorSpec(shape=(), dtype=_TYPE_DICT[dtype])

    if isinstance(data, (sw_type.BaseArtifact, sw_type.BoundingBox)):
        return tf.TensorSpec(shape=data.shape, dtype=data.dtype)
    elif isinstance(data, (list, tuple)):
        narray = numpy.array(data)
        narray_shape: t.List[t.Optional[int]] = list(narray.shape)
        narray_dtype: numpy.dtype = narray.dtype

        if narray_dtype == numpy.object_:
            _ravel_list = narray.ravel().tolist()
            _type_cnt = len(set([type(n) for n in _ravel_list]))

            if isinstance(_ravel_list[0], (list, tuple)):
                raise ValueError(f"Can't ravel to one dimension array: {_ravel_list}")

            if _type_cnt > 1:
                raise NoSupportError(
                    f"Can't convert different types in one array to tensor: {data}"
                )
            elif _type_cnt == 0:
                raise ValueError(
                    f"Can't find any types, but numpy array type is numpy.object_: {data}"
                )
            else:
                _spec = _inspect_spec(_ravel_list[0])
                if isinstance(_spec, tf.TensorSpec):
                    narray_dtype = _spec.dtype
                    _shape = _spec.shape.as_list()
                    narray_shape.extend(_shape)
                else:
                    raise NoSupportError(
                        f"Can't handle the compound type: {_ravel_list[0]}"
                    )
        return tf.TensorSpec(shape=narray_shape, dtype=narray_dtype)
    elif isinstance(data, dict):
        return {k: _inspect_spec(v) for k, v in data.items()}
    else:
        raise NoSupportError(
            f"inspect tensor spec does not support {dtype} type: {data}"
        )


def to_tf_dataset(dataset: Dataset, drop_index: bool = True) -> tf.data.Dataset:
    def _generator() -> t.Generator:
        for row in dataset:
            if drop_index:
                yield _transform(row.features)
            else:
                yield _transform(row.index), _transform(row.features)

    def _make_signature() -> t.Any:
        row = dataset.fetch_one()
        signature = []
        if not drop_index:
            signature.append(_inspect_spec(row.index))
            return _inspect_spec(row.index), _inspect_spec(row.features)
        return _inspect_spec(row.features)

    return tf.data.Dataset.from_generator(
        _generator, output_signature=_make_signature()
    )



import typing as t

import click

_TCallCommand = t.Callable[[t.Callable[..., t.Any]], click.Command]
_TCallGroup = t.Callable[[t.Callable[..., t.Any]], click.Group]
_TCallDeco = t.Union[_TCallGroup, _TCallCommand]


class AliasedGroup(click.Group):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._commands: t.Dict[str, t.List[str]] = {}
        self._aliases: t.Dict[str, str] = {}

    def _attach_aliases_args(
        self, orig_deco: _TCallDeco, aliases: t.List[str]
    ) -> _TCallDeco:
        if not aliases:
            return orig_deco

        def alias_cmd_deco(f: t.Callable) -> t.Union[click.Command, click.Group]:
            cmd = orig_deco(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return alias_cmd_deco

    def command(self, *args: t.Any, **kwargs: t.Any) -> _TCallCommand:
        aliases = kwargs.pop("aliases", [])
        cmd_deco = super().command(*args, **kwargs)
        return self._attach_aliases_args(cmd_deco, aliases)

    def group(self, *args: t.Any, **kwargs: t.Any) -> _TCallGroup:
        aliases = kwargs.pop("aliases", [])
        group_deco = super().group(*args, **kwargs)
        return self._attach_aliases_args(group_deco, aliases)  # type: ignore

    def add_command(
        self,
        cmd: click.Command,
        name: t.Optional[str] = None,
        aliases: t.Optional[t.List[str]] = None,
    ) -> None:
        aliases = aliases or []
        if aliases:
            name = name or cmd.name
            self._commands[name] = aliases
            for alias in aliases:
                self._aliases[alias] = name

        super().add_command(cmd, name)

    def get_command(
        self, ctx: click.Context, cmd_name: str
    ) -> t.Optional[click.Command]:
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        if cmd_name in self._aliases:
            return click.Group.get_command(self, ctx, self._aliases[cmd_name])

        matches = [
            x for x in self.list_commands(ctx) if x.lower().startswith(cmd_name.lower())
        ]

        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        else:
            ctx.fail(f"Too many commands matches: {', '.join(matches)}")

    def resolve_command(
        self, ctx: click.Context, args: t.List[str]
    ) -> t.Tuple[str, click.Command, t.List[str]]:
        _, cmd, args = super().resolve_command(ctx, args)
        # return command name, not alias
        return cmd.name, cmd, args

    def format_commands(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if cmd is None or cmd.hidden:
                continue

            if subcommand in self._commands:
                aliases = ",".join(sorted(self._commands[subcommand]))
                subcommand = f"{subcommand} ({aliases})"
            commands.append((subcommand, cmd))

        if commands:
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)  # type: ignore
            rows = [
                (subcommand, cmd.get_short_help_str(limit))
                for subcommand, cmd in commands
            ]
            with formatter.section("Commands"):
                formatter.write_dl(rows)



import os
import typing as t
from pathlib import Path
from urllib.parse import urlparse

from starwhale.utils import validate_obj_name
from starwhale.consts import UserRoleType, SW_API_VERSION, VERSION_PREFIX_CNT
from starwhale.base.type import URIType
from starwhale.consts.env import SWEnv
from starwhale.utils.error import URIFormatError
from starwhale.utils.config import SWCliConfigMixed

from .type import InstanceType


class ObjField:
    def __init__(
        self, typ: str = URIType.UNKNOWN, name: str = "", version: str = ""
    ) -> None:
        self.typ = typ
        self.name = name
        self.version = version

    def __str__(self) -> str:
        return f"{self.typ}-{self.name}-{self.version}"


class InstanceField(t.NamedTuple):
    uri: str
    typ: str
    alias: str


class URI:
    def __init__(self, raw: str, expected_type: str = URIType.UNKNOWN) -> None:
        self.raw = raw.strip().strip("/").lower()
        self._sw_config = SWCliConfigMixed()

        self.expected_type = expected_type
        self.full_uri = self.raw

        self.instance = ""
        self.instance_type = ""
        self.instance_alias = ""
        self._owner = ""
        self._project = ""
        self.object = ObjField()

        self._do_parse()

    def __str__(self) -> str:
        return self.full_uri

    def __repr__(self) -> str:
        return f"instance:{self.instance}, instance_type:{self.instance_type}, project:{self.project}, object:{self.object}"

    def _do_parse_instance_uri(self, raw: str) -> t.Tuple[InstanceField, str]:
        _remain: str = raw

        if not raw:
            return (
                InstanceField(
                    uri=self._sw_config._current_instance_obj["uri"],
                    typ=self._sw_config._current_instance_obj["type"],
                    alias=self._sw_config.current_instance,
                ),
                "",
            )

        _inst: str = ""
        _inst_type: str = ""
        _inst_alias: str = ""

        if raw.startswith(("http://", "https://", "cloud://")):
            _up = urlparse(raw)
            if raw.startswith("cloud://"):
                _inst = (
                    self._sw_config._config["instances"]
                    .get(_up.netloc, {})
                    .get("uri", "")
                )
            else:
                _inst = f"{_up.scheme}://{_up.netloc}"
            _remain = _up.path
            _inst_type = InstanceType.CLOUD
            _inst_alias = self._sw_config._get_instance_alias(_inst)
        elif raw.startswith("local/") or raw == "local":
            _inst = _inst_alias = "local"
            _sp = raw.split("local/", 1)
            _remain = "" if len(_sp) == 1 else _sp[1]
            _inst_type = InstanceType.STANDALONE
        else:
            _sp = raw.split("/", 1)
            _inst_alias = _sp[0]
            _inst = (
                self._sw_config._config["instances"].get(_inst_alias, {}).get("uri", "")
            )
            _inst_type = InstanceType.CLOUD
            if _inst:
                _remain = "" if len(_sp) == 1 else _sp[1]

        if not _inst:
            _inst = self._sw_config._current_instance_obj["uri"]
            _inst_type = self._sw_config._current_instance_obj["type"]
            _inst_alias = self._sw_config.current_instance

        if self.expected_type == URIType.INSTANCE:
            _remain = ""

        return InstanceField(_inst, _inst_type, _inst_alias), _remain

    def _do_parse_project_uri(self, raw: str) -> t.Tuple[str, str]:
        raw = raw.strip().strip("/")
        if not raw:
            return self._sw_config.current_project, ""

        if self.expected_type == URIType.PROJECT and not raw.startswith(
            URIType.PROJECT + "/"
        ):
            project, owner = URI.uri_to_project_and_owner(raw)
            ok, reason = validate_obj_name(project)
            if not ok:
                raise Exception(reason)
            return raw, ""

        if not raw:
            _proj = self._sw_config.current_project
            _remain = ""
        else:
            _sp = raw.split("/", 2)
            if _sp[0] == "project" and len(_sp) > 1:
                _proj = _sp[1]
                _remain = _sp[2] if len(_sp) == 3 else ""
            else:
                _proj = self._sw_config.current_project
                _remain = raw

        return _proj, _remain

    def _do_parse_obj_uri(self, raw: str) -> t.Tuple[ObjField, str]:
        raw = raw.strip().strip("/")
        if not raw:
            return ObjField(), raw

        _all_types = (
            URIType.DATASET,
            URIType.INSTANCE,
            URIType.MODEL,
            URIType.EVALUATION,
            URIType.RUNTIME,
        )

        if self.expected_type in _all_types and not raw.startswith(
            self.expected_type + "/"
        ):
            if "/version/" in raw:
                _name, _version = raw.split("/version/")
            else:
                _name, _version = raw, ""

            return ObjField(typ=self.expected_type, name=_name, version=_version), ""

        _sp = raw.split("/")
        if _sp[0] not in _all_types or len(_sp) < 2:
            raise URIFormatError(f"raw: {self.raw}, failed to parse object")

        if len(_sp) >= 4 and _sp[2] == "version":
            _version = _sp[3]
            _remain = "/".join(_sp[4:])
        else:
            _version = ""
            _remain = "/".join(_sp[2:])

        return ObjField(typ=_sp[0], name=_sp[1], version=_version), _remain

    def _do_parse(self) -> None:
        _inst, _remain = self._do_parse_instance_uri(self.raw)
        _proj, _remain = self._do_parse_project_uri(_remain)
        _obj, _remain = self._do_parse_obj_uri(_remain)

        self.instance = _inst.uri
        self.instance_type = _inst.typ
        self.instance_alias = _inst.alias
        self.project = _proj
        self.object = _obj

        self.full_uri = _inst.uri
        if _proj:
            self.full_uri = f"{self.full_uri}/project/{_proj}"

            if _obj.name:
                self.full_uri = f"{self.full_uri}/{_obj.typ}/{_obj.name}"

                if _obj.version:
                    self.full_uri = f"{self.full_uri}/version/{_obj.version}"

    def asdict(self) -> t.Dict:
        return {"full_uri": self.full_uri}

    @property
    def real_request_uri(self) -> t.Union[str, Path]:
        if self.instance_type == InstanceType.CLOUD:
            _up = urlparse(self.full_uri)
            return f"{_up.scheme}://{_up.netloc}/api/{SW_API_VERSION}{_up.path}"
        else:
            rt = self._sw_config.rootdir
            if self.project:
                rt = rt / self.project

                if self.object.name:
                    rt = rt / self.object.typ / self.object.name

                    if self.object.version:
                        rt = (
                            rt
                            / self.object.version[:VERSION_PREFIX_CNT]
                            / self.object.version
                        )

            return rt

    @property
    def sw_instance_config(self) -> t.Dict[str, t.Any]:
        return self._sw_config.get_sw_instance_config(self.instance)

    @property
    def sw_token(self) -> str:
        return self.sw_instance_config.get("sw_token", "") or os.environ.get(
            SWEnv.instance_token, ""
        )

    @property
    def sw_remote_addr(self) -> str:
        return self.sw_instance_config.get("uri", "")

    @property
    def user_name(self) -> str:
        return self.sw_instance_config.get("user_name", "")

    @property
    def user_role(self) -> str:
        return self.sw_instance_config.get("user_role", UserRoleType.NORMAL)

    @property
    def project(self) -> str:
        return URI.project_and_owner_to_uri(self._project, self._owner)

    @project.setter
    def project(self, proj: str) -> None:
        self._project, self._owner = URI.uri_to_project_and_owner(proj)

    @staticmethod
    def project_and_owner_to_uri(proj: str, owner: str) -> str:
        if owner:
            return f"{owner}:{proj}"
        else:
            return proj

    @staticmethod
    def uri_to_project_and_owner(proj: str) -> t.Tuple[str, str]:
        _op = proj.split(":", 1)
        if len(_op) > 1:
            return _op[1], _op[0]
        else:
            return _op[0], ""

    @classmethod
    def capsulate_uri_str(
        cls,
        instance: str,
        project: str = "",
        obj_type: str = "",
        obj_name: str = "",
        obj_ver: str = "",
    ) -> str:
        _fmt: t.Callable[[str], str] = lambda x: x.strip().strip("/").lower()

        _rt = f"{_fmt(instance)}"
        if project:
            _rt = f"{_rt}/project/{project}"
        else:
            return _rt

        if obj_name and obj_type:
            _rt = f"{_rt}/{obj_type}/{obj_name}"
        else:
            return _rt

        if obj_ver:
            _rt = f"{_rt}/version/{obj_ver}"

        return _rt

    @classmethod
    def capsulate_uri(
        cls,
        instance: str,
        project: str = "",
        obj_type: str = "",
        obj_name: str = "",
        obj_ver: str = "",
    ) -> "URI":
        _uri = cls.capsulate_uri_str(instance, project, obj_type, obj_name, obj_ver)
        return cls(_uri)



from typing import Any, Optional
from dataclasses import dataclass

from starwhale.base.uri import URI
from starwhale.base.uricomponents.instance import Instance
from starwhale.base.uricomponents.exceptions import UriTooShortException


@dataclass
class Project:
    name: str
    instance: Instance
    path: str = ""

    def __init__(
        self,
        name: str = "",
        uri: Optional[str] = None,
        instance: Optional[Instance] = None,
        **kwargs: Any,
    ) -> None:
        if name and uri:
            raise Exception("name and uri can not both set")
        # init instance
        if "/" in name:
            uri = name
            name = ""
        self.instance = instance or Instance(uri=uri or "", **kwargs)

        if not name:
            # use project name in path, the path must at least contain project/{name}
            p = self.instance.path.split("/")
            if len(p) >= 2 and p[0] == "project":
                name = p[1]
                self.path = "/".join(p[2:])
            elif len(p) == 1:
                # Compatible with features of URI
                name = p[0]

        if not name:
            # use default project
            name = self.instance.info.get("current_project", "")
        if not name:
            raise Exception("can not init project with empty name")
        self.name = name
        # TODO check if project exists for local and remote

    @classmethod
    def parse_from_full_uri(cls, uri: str, ignore_rc_type: bool) -> "Project":
        """
        Parse project from full uri.
        we do not parse instance and project info from uri less than 5 parts.
        we prefer that users use `dataset copy mnist -dlp project` rather than `dataset copy project/dataset/mnist`.
        the second is difficult to write correctly at once and the semantics are not very clear.
        and the long uri usually copied from the website.
        """
        if "://" in uri:
            no_schema_uri = uri.split("://", 1)[-1]
        else:
            no_schema_uri = uri

        if "//" in no_schema_uri:
            raise Exception(f"wrong format uri({uri}) with '//'")

        parts = no_schema_uri.split("/")
        exp = len("local/project/self/dataset/mnist".split("/"))
        if ignore_rc_type:
            # ignore type in uri like dataset
            exp = exp - 1

        if len(parts) < exp:
            raise UriTooShortException(
                exp, len(parts), f"can not parse project info from {uri}"
            )
        return cls(uri=uri)

    def to_uri(self) -> URI:
        return URI.capsulate_uri(instance=str(self.instance), project=self.name)



import os
import typing as t
import getpass as gt
from pwd import getpwnam
from pathlib import Path

from loguru import logger

from starwhale.utils import console, now_str, is_darwin, gen_uniq_version
from starwhale.consts import CNTR_DEFAULT_PIP_CACHE_DIR
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir
from starwhale.base.type import URIType, EvalTaskType, RunSubDirType
from starwhale.consts.env import SWEnv
from starwhale.utils.error import NoSupportError, FieldTypeOrValueError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.utils.process import check_call
from starwhale.utils.progress import run_with_progress_bar
from starwhale.core.eval.store import EvaluationStorage
from starwhale.core.model.model import StandaloneModel
from starwhale.core.runtime.model import StandaloneRuntime


class EvalExecutor:
    def __init__(
        self,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        project_uri: URI,
        version: str = "",
        name: str = "",
        job_name: str = "default",
        desc: str = "",
        step: str = "",
        task_index: int = 0,
        task_num: int = 0,
        gencmd: bool = False,
        use_docker: bool = False,
    ) -> None:
        self.name = name
        self.job_name = job_name

        if step:
            self.type = EvalTaskType.SINGLE
        else:
            self.type = EvalTaskType.ALL
        self.step = step
        self.task_index = task_index
        self.task_num = task_num

        self.desc = desc
        self.model_uri = model_uri
        self.project_uri = project_uri
        self.dataset_uris = [
            URI(u, expected_type=URIType.DATASET) for u in dataset_uris
        ]

        self.runtime_uri = runtime_uri

        if runtime_uri:
            self.runtime: t.Optional[StandaloneRuntime] = StandaloneRuntime(
                URI(runtime_uri, expected_type=URIType.RUNTIME)
            )
            self.baseimage = self.runtime.store.get_docker_base_image()
        else:
            self.runtime = None
            self.baseimage = ""

        self.project_dir = Path(self.project_uri.real_request_uri)

        self.gencmd = gencmd
        self.use_docker = use_docker

        if not version:
            logger.info("[step:init]create eval job version...")
            self._version = gen_uniq_version()
            logger.info(f"[step:init]eval job version is {self._version}")
        else:
            self._version = version

        self.sw_config = SWCliConfigMixed()

        self._workdir = Path()

        _m = StandaloneModel(URI(self.model_uri, expected_type=URIType.MODEL))
        self._src_dir = _m.store.src_dir
        self._runtime_dir = Path()

        self._do_validate()

    def _do_validate(self) -> None:
        if self.use_docker:
            if not self.runtime_uri:
                raise FieldTypeOrValueError("runtime_uri is none")
            if is_darwin(arm=True):
                raise NoSupportError(
                    "use docker as the evaluation job environment in macOS system (Apple Silicon processor)"
                )

    def __str__(self) -> str:
        return f"Evaluation Executor: {self.name}"

    def __repr__(self) -> str:
        return f"Evaluation Executor: name -> {self.name}, version -> {self._version}"

    def run(self) -> str:
        try:
            self._do_run()
        except Exception as e:
            logger.error(f"execute evaluation error, error is:{str(e)}")
            raise e
        return self._version

    def _do_run(self) -> None:
        operations = [
            (self._prepare_workdir, 5, "prepare workdir"),
            (self._extract_swrt, 15, "extract runtime"),
            (self._do_run_cmd, 70, "run eval job"),
        ]

        run_with_progress_bar("eval run in local...", operations)

    def _prepare_workdir(self) -> None:
        logger.info("[step:prepare]create eval workdir...")
        # TODO: fix _workdir sequence-dependency issue
        self._workdir = EvaluationStorage.local_run_dir(
            self.project_uri.project, self._version
        )

        ensure_dir(self._workdir)
        ensure_dir(self._workdir / RunSubDirType.SWMP)

        logger.info(f"[step:prepare]eval workdir: {self._workdir}")

    def _extract_swrt(self) -> None:
        if self.runtime and self.use_docker:
            # avoid conflict with normal process with venv or conda
            self._runtime_dir = self.runtime.extract()
        else:
            self._runtime_dir = Path()

    def _do_run_cmd(self) -> None:
        if self.use_docker:
            self._do_run_cmd_in_container()
        else:
            self._do_run_cmd_in_host()

    def _do_run_cmd_in_host(self) -> None:
        StandaloneModel.eval_user_handler(
            project=self.project_uri.project,
            version=self._version,
            workdir=self._src_dir,
            dataset_uris=[u.full_uri for u in self.dataset_uris],
            step_name=self.step,
            task_index=self.task_index,
            task_num=self.task_num,
            # other runtime info
            base_info=dict(
                name=self.name,
                desc=self.desc,
                model=self.model_uri,
                model_dir=str(self._src_dir),
                datasets=[u.full_uri for u in self.dataset_uris],
                runtime=self.runtime_uri,
                created_at=now_str(),
            ),
        )

    def _do_run_cmd_in_container(self) -> None:
        cmd = self._gen_run_container_cmd(self.type, self.step, self.task_index)
        console.rule(f":elephant: {self.type} docker cmd", align="left")
        console.print(f"{cmd}\n")
        console.print(
            f":fish: eval run:{self.type} dir @ [green blink]{self._workdir}/{self.type}[/]"
        )
        if not self.gencmd:
            check_call(f"docker pull {self.baseimage}", shell=True)
            check_call(cmd, shell=True)

    def _gen_run_container_cmd(self, typ: str, step: str, task_index: int) -> str:
        if typ not in (EvalTaskType.ALL, EvalTaskType.SINGLE):
            raise Exception(f"no support {typ} to gen docker cmd")
        _entrypoint = "run"
        _run_dir = self._workdir

        cmd = [
            "docker",
            "run",
            "--net=host",
            "--rm",
            "--name",
            f"{self._version}-{step}-{task_index}",
            "-e",
            "DEBUG=1",
            "-e",
            f"SW_USER={gt.getuser()}",
            "-e",
            f"SW_USER_ID={getpwnam(gt.getuser()).pw_uid}",
            "-e",
            "SW_USER_GROUP_ID=0",
            "-e",
            f"SW_LOCAL_STORAGE={self.sw_config.rootdir}",
            "-l",
            f"version={self._version}",
        ]

        cmd += [
            "-v",
            f"{_run_dir}:{_run_dir}",
            "-v",
            f"{self.sw_config.rootdir}:{self.sw_config.rootdir}",
            "-v",
            f"{self.sw_config.object_store_dir}:{self.sw_config.object_store_dir}",
        ]

        if typ == EvalTaskType.SINGLE:
            cmd.extend(["-e", f"SW_TASK_STEP={step}"])
            cmd.extend(["-e", f"SW_TASK_INDEX={task_index}"])

        logger.debug(f"config:{self.sw_config._current_instance_obj}")

        cmd.extend(["-e", f"{SWEnv.project}={self.project_uri.project}"])
        cmd.extend(["-e", f"{SWEnv.eval_version}={self._version}"])
        cmd.extend(["-e", f"{SWEnv.model_version}={self.model_uri}"])
        cmd.extend(["-e", f"{SWEnv.runtime_version}={self.runtime_uri}"])
        cmd.extend(
            [
                "-e",
                f"{SWEnv.instance_uri}={self.sw_config._current_instance_obj['uri']}",
                "-e",
                f"{SWEnv.instance_token}={self.sw_config._current_instance_obj.get('sw_token', '')}",
            ]
        )
        # support multi dataset
        cmd.extend(
            [
                "-e",
                f"{SWEnv.dataset_uri}={' '.join([ds.full_uri for ds in self.dataset_uris])}",
            ]
        )

        cntr_cache_dir = os.environ.get("SW_PIP_CACHE_DIR", CNTR_DEFAULT_PIP_CACHE_DIR)
        host_cache_dir = os.path.expanduser("~/.cache/starwhale-pip")
        cmd += ["-v", f"{host_cache_dir}:{cntr_cache_dir}"]

        _env = os.environ
        for _ee in (
            "SW_PYPI_INDEX_URL",
            "SW_PYPI_EXTRA_INDEX_URL",
            "SW_PYPI_TRUSTED_HOST",
        ):
            if _ee not in _env:
                continue
            cmd.extend(["-e", f"{_ee}={_env[_ee]}"])

        cmd += [self.baseimage, _entrypoint]
        return " ".join(cmd)



import click

from starwhale.mngt import gc, check, open_web


def add_mngt_command(cli: click.core.Group) -> None:
    @cli.command("gc", help="Standalone garbage collection")
    @click.option("--dry-run", is_flag=True, help="Dry-run cleanup garbage collection")
    @click.option("--yes", is_flag=True, help="All confirm yes")
    def _gc(dry_run: bool, yes: bool) -> None:
        gc(dry_run, yes)

    @cli.command("ui", help="Open instance web ui")
    @click.argument("instance", default="")
    def _ui(instance: str) -> None:
        open_web(instance)

    @cli.command("check", help="Check external dependency software")
    def _check() -> None:
        check()



import time
import random

import click

from starwhale.version import STARWHALE_VERSION
from starwhale.utils.cli import AliasedGroup
from starwhale.utils.debug import init_logger
from starwhale.utils.config import load_swcli_config
from starwhale.core.eval.cli import eval_job_cmd
from starwhale.core.model.cli import model_cmd
from starwhale.cli.board.board import open_board
from starwhale.core.dataset.cli import dataset_cmd
from starwhale.core.project.cli import project_cmd
from starwhale.core.runtime.cli import runtime_cmd
from starwhale.core.instance.cli import instance_cmd
from starwhale.cli.assistance.cli import assistance_cmd

from .mngt import add_mngt_command
from .config import config_cmd
from .completion import completion_cmd


def create_sw_cli() -> click.core.Group:
    @click.group(cls=AliasedGroup)
    @click.version_option(version=STARWHALE_VERSION, message="%(version)s")
    @click.option(
        "-v",
        "--verbose",
        count=True,
        help="Show verbose log, support multi counts for v args. More v args, more logs.",
    )
    @click.option("-o", "--output", help="Output format", type=click.Choice(["json"]))
    @click.pass_context
    def cli(ctx: click.Context, verbose: bool, output: str) -> None:
        load_swcli_config()
        init_logger(verbose)
        ctx.ensure_object(dict)
        ctx.obj["output"] = output

    random.seed(time.time_ns())

    cli.add_command(instance_cmd)
    cli.add_command(project_cmd, aliases=["prj"])  # type: ignore
    cli.add_command(eval_job_cmd)
    cli.add_command(runtime_cmd, aliases=["rt"])  # type: ignore
    cli.add_command(model_cmd, aliases=["mp"])  # type: ignore
    cli.add_command(dataset_cmd, aliases=["ds"])  # type: ignore
    cli.add_command(open_board)
    cli.add_command(completion_cmd)
    cli.add_command(config_cmd)
    cli.add_command(assistance_cmd)
    add_mngt_command(cli)

    return cli


cli = create_sw_cli()

if __name__ == "__main__":
    cli()



import os
import time
import random
import string
import threading
import traceback
from typing import IO, Any, List, Union, TextIO, Callable, Optional
from datetime import datetime

import requests


def random_id() -> str:
    return "".join(
        [random.choice(string.ascii_letters + string.digits) for i in range(8)]
    )


def check_status_code(resp: requests.Response) -> None:
    if resp.status_code != 200:
        msg = f"status code={resp.status_code} body={resp.json()}"
        if resp.status_code // 100 == 4:
            raise UnrecoverableError(msg)
        raise RuntimeError(msg)


class UnrecoverableError(Exception):
    pass


class InvalidOffsetError(Exception):
    pass


class ClosedError(Exception):
    pass


class TaskRunner(threading.Thread):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)
        self.stopped = False
        self.daemon = True

    def stop(self) -> None:
        self.stopped = True

    def run_until_success(
        self, f: Callable[[], Any], ignore_stopped: bool = False
    ) -> Any:
        backoff = 0.1
        while not self.stopped or ignore_stopped:
            try:
                return f()
            except UnrecoverableError:
                raise
            except Exception:
                print("===Retry due to exception===")
                traceback.print_exc()
                print("======")
            time.sleep(backoff)
            backoff *= 2
            if backoff > 1:
                backoff = 1


class PeriodicRunner(TaskRunner):
    def __init__(self, name: str, f: Callable[[], None], interval: float):
        super().__init__(name)
        self.f = f
        self.interval = interval

    def run(self) -> None:
        while not self.stopped:
            try:
                start_time = datetime.now().timestamp()
                while not self.stopped:
                    now = datetime.now().timestamp()
                    elasped = now - start_time
                    if elasped > self.interval:
                        break
                    time.sleep(min(1, self.interval - elasped))
                if not self.stopped:
                    self.f()
            except Exception:
                traceback.print_exc()


class ChunkBuffer:
    def __init__(self, chunk_size: int, max_chunk_count: int):
        self.chunk_size = chunk_size
        self.max_chunk_count = max_chunk_count
        self.chunks: List[bytes] = []
        self.offset = 0
        self.length = 0
        self.lock = threading.Condition()
        self.error: Optional[str] = None

    def write(self, offset: int, data: bytes, timeout: float) -> int:
        with self.lock:
            if self.error is not None:
                raise ClosedError(self.error)
            if offset > self.length:
                raise InvalidOffsetError(f"invalid offset {offset}")
            skip_len = self.length - offset
            if skip_len >= len(data):
                return len(data)
            data = data[skip_len:]
            index = 0
            while index < len(data):
                if self.error is not None:
                    raise ClosedError(self.error)
                if len(self.chunks) == 0 or len(self.chunks[-1]) == self.chunk_size:
                    if len(self.chunks) == self.max_chunk_count:
                        if index == 0 and self.lock.wait(timeout):
                            continue
                        break
                    self.chunks.append(b"")
                    available = self.chunk_size
                else:
                    available = self.chunk_size - len(self.chunks[-1])
                if available > len(data) - index:
                    available = len(data) - index
                self.chunks[-1] += data[index : index + available]
                index += available
            self.lock.notify_all()
            self.length += index
            return index + skip_len

    def read(self, offset: int, timeout: float) -> Optional[bytes]:
        with self.lock:
            if offset < self.offset or offset > self.length:
                raise InvalidOffsetError(f"invalid offset {offset}")
            while self.error is None and offset == self.length:
                if not self.lock.wait(timeout):
                    return None
            if self.error is not None and self.error != "":
                raise ClosedError(self.error)
            if offset == self.length:
                return b""
            index = 0
            while self.offset + len(self.chunks[index]) <= offset:
                self.offset += len(self.chunks[index])
                index += 1
            if index > 0:
                self.chunks = self.chunks[index:]
                self.lock.notify_all()
            return self.chunks[0][offset - self.offset :]

    def close(self, error: str) -> None:
        with self.lock:
            if self.error is None:
                self.error = error
                self.lock.notify_all()


class ChunkBufferReader:
    def __init__(self, chunk_buffer: ChunkBuffer):
        self.chunk_buffer = chunk_buffer
        self.offset = 0

    def read(self, timeout: float) -> Optional[bytes]:
        data = self.chunk_buffer.read(self.offset, timeout)
        if data is not None:
            self.offset += len(data)
        return data

    def close(self, error: str) -> None:
        self.chunk_buffer.close(error)


class ChunkBufferWriter:
    def __init__(self, chunk_buffer: ChunkBuffer):
        self.chunk_buffer = chunk_buffer
        self.offset = 0

    def write(self, data: bytes, timeout: float) -> int:
        count = self.chunk_buffer.write(self.offset, data, timeout)
        self.offset += count
        return count

    def close(self, error: str) -> None:
        self.chunk_buffer.close(error)


class FileReader(TaskRunner):
    def __init__(
        self,
        name: str,
        file: Union[TextIO, IO[bytes]],
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.file = file
        self.chunk_buffer_writer = ChunkBufferWriter(chunk_buffer)

    def run(self) -> None:
        try:
            fd = self.file.fileno()
            while not self.stopped:
                data = os.read(fd, 16384)
                if len(data) == 0:
                    self.chunk_buffer_writer.close("")
                    return
                while not self.stopped and len(data) > 0:
                    count = self.chunk_buffer_writer.write(data, 0.5)
                    data = data[count:]
        except Exception as e:
            print("===Stop read from file===")
            traceback.print_exc()
            print("======")
            self.chunk_buffer_writer.close(str(e))
        finally:
            self.file.close()


class FileWriter(TaskRunner):
    def __init__(
        self,
        name: str,
        file: Union[TextIO, IO[bytes]],
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.file = file
        self.chunk_buffer_reader = ChunkBufferReader(chunk_buffer)
        self.offset = 0

    def run(self) -> None:
        try:
            fd = self.file.fileno()
            while not self.stopped:
                data = self.chunk_buffer_reader.read(0.5)
                if data is not None:
                    if self.stopped or len(data) == 0:
                        return
                    self.offset += len(data)
                    while not self.stopped and len(data) > 0:
                        count = os.write(fd, data)
                        data = data[count:]
        except Exception as e:
            print("===Stop write to file===")
            traceback.print_exc()
            print("======")
            self.chunk_buffer_reader.close(str(e))
        finally:
            self.file.close()


class BrokerReader(TaskRunner):
    def __init__(
        self,
        name: str,
        broker_url: str,
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.broker_url = broker_url
        self.chunk_buffer_writer = ChunkBufferWriter(chunk_buffer)
        self.offset = 0

    def run(self) -> None:
        try:
            while not self.stopped:
                data = self.run_until_success(self._read_from_broker)
                if data is not None:
                    if len(data) == 0:
                        self.chunk_buffer_writer.close("")
                        return
                    self.offset += len(data)
                    while not self.stopped and len(data) > 0:
                        count = self.chunk_buffer_writer.write(data, 0.5)
                        data = data[count:]
        except Exception as e:
            print("===Stop read from broker===")
            traceback.print_exc()
            print("======")
            self.chunk_buffer_writer.close(str(e))
        finally:
            self.run_until_success(self._close_read, True)

    def _read_from_broker(self) -> Optional[bytes]:
        resp = requests.get(self.broker_url, params={"offset": self.offset}, timeout=5)
        check_status_code(resp)
        if "x-wait-for-data" in resp.headers:
            return None
        return resp.content

    def _close_read(self) -> None:
        resp = requests.post(f"{self.broker_url}/closeRead", timeout=5)
        check_status_code(resp)


class BrokerWriter(TaskRunner):
    def __init__(
        self,
        name: str,
        broker_url: str,
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.broker_url = broker_url
        self.chunk_buffer_reader = ChunkBufferReader(chunk_buffer)
        self.offset = 0

    def run(self) -> None:
        try:
            while not self.stopped:
                data = self.chunk_buffer_reader.read(0.5)
                if data is not None:
                    if self.stopped or len(data) == 0:
                        return
                    d = data
                    while not self.stopped and len(d) > 0:
                        count = self.run_until_success(lambda: self._write_to_broker(d))
                        self.offset += count
                        d = d[count:]
        except Exception as e:
            print("===Stop write to broker===")
            traceback.print_exc()
            print("======")
            self.chunk_buffer_reader.close(str(e))
        finally:
            self.run_until_success(self._close_write, True)

    def _write_to_broker(self, data: bytes) -> int:
        resp = requests.post(
            self.broker_url, params={"offset": self.offset}, data=data, timeout=5
        )
        check_status_code(resp)
        return resp.json()["count"]  # type: ignore

    def _close_write(self) -> None:
        resp = requests.post(f"{self.broker_url}/closeWrite", timeout=5)
        check_status_code(resp)



from __future__ import annotations

import json
import queue
import typing as t
import threading
from pathlib import Path
from collections import defaultdict

from starwhale.consts import VERSION_PREFIX_CNT
from starwhale.utils.fs import (
    ensure_dir,
    ensure_file,
    blake2b_content,
    BLAKE2B_SIGNATURE_ALGO,
)
from starwhale.utils.error import NoSupportError
from starwhale.core.dataset.type import Link, BaseArtifact
from starwhale.api._impl.data_store import TableWriter, LocalDataStore

from .base import (
    TrackRecord,
    ParamsRecord,
    MetricsRecord,
    ArtifactsRecord,
    MetricsTabularRow,
    ArtifactsTabularRow,
    HandleQueueElementType,
)


class HandlerThread(threading.Thread):
    def __init__(
        self,
        workdir: t.Union[Path, str],
        handle_queue: queue.Queue[HandleQueueElementType],
        sync_queue: t.Optional[queue.Queue[t.Any]] = None,
    ) -> None:
        super().__init__(name="HandlerThread")

        self.handle_queue = handle_queue
        self.sync_queue = sync_queue

        self._run_exception: t.Optional[Exception] = None
        self._workdir = Path(workdir)
        self._data_store = LocalDataStore(str(self._workdir))
        self._table_writers: t.Dict[str, TableWriter] = {}
        self._params: t.Dict[str, t.Dict] = defaultdict(dict)
        # TODO: support non-in-memory artifacts auto-incr counter with datastore
        self._artifacts_counter: t.Dict = defaultdict(int)

        self.daemon = True

    def _raise_run_exception(self) -> None:
        if self._run_exception is not None:
            raise threading.ThreadError(
                f"HandlerThread raise exception: {self._run_exception}"
            )

    def flush(self) -> None:
        self._dump_params()

        for writer in self._table_writers.values():
            writer.flush()

        self._data_store.dump()

    def close(self) -> None:
        self.handle_queue.put(None)
        self.join()
        self.flush()

        for writer in self._table_writers.values():
            writer.close()

        self._raise_run_exception()

    def _handle_metrics(self, record: MetricsRecord) -> None:
        row = MetricsTabularRow.from_record(record)
        table_name = f"{record.typ.value}/{record.source.value}"
        self._update_table(table_name, row.asdict(), "__step")

    def _handle_artifacts(self, record: ArtifactsRecord) -> None:
        table_name = f"{record.typ.value}/{record.source.value}"
        store_dir = self._workdir / record.typ.value / "_files"

        def _convert_data_to_link(data: BaseArtifact) -> Link:
            raw_content = data.to_bytes()
            sign_name = blake2b_content(raw_content)
            fpath = (
                store_dir
                / BLAKE2B_SIGNATURE_ALGO
                / sign_name[:VERSION_PREFIX_CNT]
                / sign_name
            )
            ensure_file(fpath, raw_content, parents=True)

            return Link(
                uri=sign_name,
                offset=0,
                size=len(raw_content),
                use_plain_type=True,
                data_type=data,
            )

        for k, v in record.data.items():
            if isinstance(v, Link):
                data = v
            elif isinstance(v, BaseArtifact):
                data = _convert_data_to_link(v)
            else:
                raise NoSupportError(
                    f"artifact only accepts BaseArtifact or Link type: {v}"
                )

            row = ArtifactsTabularRow(
                name=k,
                index=self._artifacts_counter[k],
                created_at=record.clock_time,
                data=data,
                link_wrapper=not isinstance(v, Link),
            )
            self._update_table(table_name, row.asdict(), "__key")
            self._artifacts_counter[k] += 1

    def _handle_params(self, record: ParamsRecord) -> None:
        self._params[record.source.value].update(record.data)

    def _dump_params(self) -> None:
        params_dir = self._workdir / "params"
        ensure_dir(params_dir)

        for k, v in self._params.items():
            if not v:
                continue
            ensure_file(params_dir / f"{k}.json", json.dumps(v, separators=(",", ":")))

    def _dispatch(self, record: TrackRecord) -> None:
        if isinstance(record, MetricsRecord):
            self._handle_metrics(record)
        elif isinstance(record, ParamsRecord):
            self._handle_params(record)
        elif isinstance(record, ArtifactsRecord):
            self._handle_artifacts(record)
        else:
            raise NoSupportError(f"no support to handle {record}({type(record)})")

    def _update_table(self, table_name: str, data: t.Dict, key_column: str) -> None:
        if table_name not in self._table_writers:
            self._table_writers[table_name] = TableWriter(
                table_name=table_name,
                key_column=key_column,
                data_store=self._data_store,
                run_exceptions_limits=10,
            )

        writer = self._table_writers[table_name]
        writer.insert(data)

    def run(self) -> None:
        try:
            while True:
                record = self.handle_queue.get(block=True, timeout=None)
                if record is None:
                    if self.handle_queue.qsize() > 0:
                        continue
                    else:
                        break  # pragma: no cover

                self._dispatch(record)
        except Exception as e:
            self._run_exception = e
            raise



from ._impl.track.tracker import Tracker

start = Tracker.start
end = Tracker.end
metrics = Tracker.metrics
artifacts = Tracker.artifacts
params = Tracker.params

__all__ = ["start", "end", "metrics", "artifacts", "params", "Tracker"]



import typing as t

from starwhale.consts import DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE, STANDALONE_INSTANCE
from starwhale.base.uri import URI
from starwhale.base.type import URIType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.http import ignore_error


class Instance:
    uri: URI


class StandaloneInstance(Instance):
    def __init__(self) -> None:
        super().__init__()
        self.uri = URI(STANDALONE_INSTANCE, expected_type=URIType.INSTANCE)


class CloudInstance(Instance, CloudRequestMixed):
    def __init__(self, uri: str) -> None:
        super().__init__()
        self.uri = URI(uri, expected_type=URIType.INSTANCE)

    @ignore_error("--")
    def _fetch_version(self) -> str:
        return self.do_http_request("/system/version", instance_uri=self.uri).json()[  # type: ignore
            "data"
        ][
            "version"
        ]

    @ignore_error([])
    def _fetch_agents(self) -> t.List[t.Dict[str, t.Any]]:
        # TODO: add pageSize to args
        return self.do_http_request(  # type: ignore
            "/system/agent",
            params={"pageSize": DEFAULT_PAGE_SIZE, "pageNum": DEFAULT_PAGE_IDX},
            instance_uri=self.uri,
        ).json()["data"]["list"]

    @ignore_error({})
    def _fetch_current_user(self) -> t.Dict[str, t.Any]:
        r = self.do_http_request("/user/current", instance_uri=self.uri).json()["data"]
        return dict(name=r["name"], role=r["role"]["roleName"])



import os
import sys
import typing as t
import subprocess
import webbrowser
from pathlib import Path

import click
from packaging.version import parse as version_parse

from starwhale.utils import console
from starwhale.consts import (
    RECOVER_DIRNAME,
    SW_TMP_DIR_NAME,
    DATA_STORE_DIRNAME,
    VERSION_PREFIX_CNT,
    STANDALONE_INSTANCE,
    OBJECT_STORE_DIRNAME,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.utils.fs import empty_dir
from starwhale.base.type import URIType, get_bundle_type_by_uri
from starwhale.utils.venv import get_conda_bin
from starwhale.utils.config import SWCliConfigMixed


def gc(dry_run: bool = False, yes: bool = False) -> None:
    sw = SWCliConfigMixed()

    for project_dir in sw.rootdir.iterdir():
        project_name = project_dir.name
        if project_name in (
            RECOVER_DIRNAME,
            SW_TMP_DIR_NAME,
            OBJECT_STORE_DIRNAME,
            DATA_STORE_DIRNAME,
        ):
            continue

        if not yes and not click.confirm(
            f"continue to garbage collection for project: {project_name}?"
        ):
            continue

        _removed_paths = []
        for typ in (
            URIType.DATASET,
            URIType.EVALUATION,
            URIType.RUNTIME,
            URIType.MODEL,
        ):
            _bundle_type = (
                get_bundle_type_by_uri(typ) if typ != URIType.EVALUATION else ""
            )
            _recover_dir = project_dir / typ / RECOVER_DIRNAME

            if typ in (URIType.RUNTIME, URIType.MODEL):
                for _path in _recover_dir.glob(f"**/*{_bundle_type}"):
                    if not (_path.is_file() and _path.name.endswith(_bundle_type)):
                        continue
                    _removed_paths.append(_path)
                    _workdir_path = _get_workdir_path(project_dir, typ, _path)
                    if _workdir_path.exists():
                        _removed_paths.append(_workdir_path)
            elif typ in (URIType.DATASET, URIType.EVALUATION):
                for _path in _recover_dir.glob(f"**/{DEFAULT_MANIFEST_NAME}"):
                    if not _path.is_file():
                        continue
                    _removed_paths.append(_path.parent)
                    _workdir_path = _get_workdir_path(project_dir, typ, _path.parent)
                    if _workdir_path.exists():
                        _removed_paths.append(_workdir_path)

        _rcnt = len(_removed_paths)
        if _rcnt == 0:
            console.print(f":deer: project:[red]{project_name}[/], no objects to gc")
            continue

        console.print(
            f":zap: project:[red]{project_name}[/], find {len(_removed_paths)} objects to cleanup..."
        )
        console.print("\n".join([f":no_entry_sign: {_p}" for _p in _removed_paths]))

        if not dry_run and (yes or click.confirm("continue to remove?")):
            for _p in _removed_paths:
                empty_dir(_p)

    _gc_special_dirs(sw.rootdir, dry_run, yes)


def _get_workdir_path(project_dir: Path, typ: str, bundle_path: Path) -> Path:
    version = bundle_path.name.split(".")[0]
    object_prefix = f"{version[:VERSION_PREFIX_CNT]}/{version}"

    if typ != URIType.EVALUATION:
        object_name = bundle_path.parent.parent.name
        object_prefix = f"{object_name}/{object_prefix}"
    _rpath = project_dir / "workdir" / typ / RECOVER_DIRNAME / object_prefix
    if _rpath.exists():
        return _rpath
    else:
        return project_dir / "workdir" / typ / object_prefix


def _gc_special_dirs(root: Path, dry_run: bool, yes: bool) -> None:
    if os.path.isdir(root / RECOVER_DIRNAME):
        for project_dir in (root / RECOVER_DIRNAME).iterdir():
            if not dry_run and (yes or click.confirm("continue to remove?")):
                empty_dir(project_dir)

            console.print(f":no_entry_sign: [red]{project_dir}[/] removed")
    if os.path.isdir(root / SW_TMP_DIR_NAME) and not dry_run:
        # no need to confirm without dry_run, they are really garbage
        empty_dir(root / SW_TMP_DIR_NAME)


def open_web(instance_uri: str = "") -> None:
    sw = SWCliConfigMixed()

    instance_uri = instance_uri.strip().strip("/")
    instance_uri = instance_uri or sw.current_instance

    if instance_uri == STANDALONE_INSTANCE:
        console.print(":see_no_evil: standalone web ui is comming soon.")
    else:
        _config = sw.get_sw_instance_config(instance_uri)
        _uri = _config.get("uri", "")
        if not _uri:
            console.print(
                f":rotating_light: sw cannot find instance [red]{instance_uri}[/], please [bold red]swcli instance login[/] first."
            )
            sys.exit(1)
        else:
            console.print(f":clap: try to open {_uri}")
            webbrowser.open(_uri)


class _CheckLevel:
    WARN = "warn"
    CRITICAL = "critical"


class _Dependency(t.NamedTuple):
    title: str
    min_version: str
    level: str
    help: str
    checker: t.Callable[..., str]


def check() -> None:
    def _check_docker() -> str:
        out = subprocess.check_output(
            ["docker", "version", "--format", "{{.Client.Version}}"],
            stderr=subprocess.STDOUT,
        )
        return out.decode().strip()

    def _check_conda() -> str:
        out = subprocess.check_output(
            [get_conda_bin(), "--version"], stderr=subprocess.STDOUT
        )
        return out.decode().strip().split()[-1]

    dependencies: t.List[_Dependency] = [
        _Dependency(
            title="Docker",
            min_version="19.03",
            level=_CheckLevel.WARN,
            help=(
                "Docker is an open platform for developing, shipping, and running applications."
                "Starwhale uses Docker to run jobs. You can visit https://docs.docker.com/get-docker/ for more details."
            ),
            checker=_check_docker,
        ),
        _Dependency(
            title="Conda",
            min_version="4.0.0",
            level=_CheckLevel.WARN,
            help=(
                "Conda is an open-source package management system and environment management system."
                "Starwhale uses Conda to build runtime. You can download it from https://docs.conda.io/en/latest/miniconda.html."
            ),
            checker=_check_conda,
        ),
    ]

    for d in dependencies:
        ok, reason = True, ""
        _check_version = ""
        try:
            _check_version = d.checker()
        except subprocess.CalledProcessError as e:
            ok, reason = False, f"exit code:{e.returncode}, command:{e.output}"
        except Exception as e:
            ok, reason = False, str(e)

        if _check_version and version_parse(_check_version) < version_parse(
            d.min_version
        ):
            ok, reason = (
                False,
                f"version: {_check_version}, expected min version: {d.min_version}",
            )

        if ok:
            console.print(f":white_check_mark: {d.title} {_check_version}")
        else:
            console.print(f":x: {d.title} [{d.level}]")
            console.print(f"\t * :point_right: Reason: [red]{reason}[/]")
            console.print(f"\t * :blue_book: Min version: {d.min_version}")
            console.print(f"\t * :information_desk_person: Advice: {d.help} \n")



import json
import os.path
from pathlib import Path

from fastapi import APIRouter
from starlette.requests import Request

from starwhale.utils import load_yaml
from starwhale.consts import (
    SW_AUTO_DIRNAME,
    EVALUATION_PANEL_LAYOUT_JSON_FILE_NAME,
    EVALUATION_PANEL_LAYOUT_YAML_FILE_NAME,
)
from starwhale.utils.fs import ensure_file
from starwhale.web.response import success, SuccessResp

router = APIRouter()
prefix = "panel"


def _setting_file() -> Path:
    base = Path(os.path.join(os.getcwd(), SW_AUTO_DIRNAME))
    yaml_fmt = base / EVALUATION_PANEL_LAYOUT_YAML_FILE_NAME
    json_fmt = base / EVALUATION_PANEL_LAYOUT_JSON_FILE_NAME
    # only one of them should exist
    # we can not decide which one to use when both exist
    if yaml_fmt.exists() and json_fmt.exists():
        raise RuntimeError("Both yaml and json file exist, please remove one of them")

    # prefer using yaml config for local development (manual edit)
    if yaml_fmt.exists():
        return yaml_fmt
    return json_fmt


def _is_yaml_fmt(file: Path) -> bool:
    return file.name == EVALUATION_PANEL_LAYOUT_YAML_FILE_NAME


@router.post("/setting/{project}/{key}")
async def setting(project: str, key: str, request: Request) -> SuccessResp:
    content = await request.body()
    f = _setting_file()
    if _is_yaml_fmt(f):
        raise RuntimeError(
            "Can not save the console layout in yaml format when the manual edit yaml is exist"
        )
    ensure_file(f, content.decode("utf-8"), parents=True)
    return success({})


@router.get("/setting/{project}/{key}")
def get_setting(project: str, key: str) -> SuccessResp:
    f = _setting_file()
    if not f.exists() or not f.is_file():
        return success("")

    if _is_yaml_fmt(f):
        # render yaml to json for console
        content = load_yaml(f)
        return success(json.dumps(content))
    with f.open("r") as fd:
        return success(fd.read())



import typing as t
from pathlib import Path

import yaml

from starwhale.utils import now_str, load_yaml, validate_obj_name
from starwhale.consts import LATEST_TAG, DEFAULT_MANIFEST_NAME
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.base.type import InstanceType
from starwhale.utils.error import (
    FormatError,
    NotFoundError,
    NoSupportError,
    MissingFieldError,
)


class StandaloneTag:
    def __init__(self, uri: URI) -> None:
        self.uri = uri
        self._do_validate()

    def _do_validate(self) -> None:
        if self.uri.instance_type != InstanceType.STANDALONE:
            raise NoSupportError(self.uri.instance_type)

    @property
    def _manifest_path(self) -> Path:
        return (
            self.uri._sw_config.rootdir
            / self.uri.project
            / self.uri.object.typ
            / self.uri.object.name
            / DEFAULT_MANIFEST_NAME
        )

    def _get_manifest(self) -> t.Dict[str, t.Any]:
        if not self._manifest_path.exists():
            _dft: t.Dict[str, t.Any] = {
                "fast_tag_seq": -1,
                "tags": {},
                "versions": {},
            }
            ensure_dir(self._manifest_path.parent)
            ensure_file(self._manifest_path, yaml.safe_dump(_dft))
            return _dft
        else:
            return load_yaml(self._manifest_path)  # type: ignore

    def _save_manifest(self, _manifest: t.Dict[str, t.Any]) -> None:
        _manifest["updated_at"] = now_str()
        _manifest["name"] = self.uri.object.name
        _manifest["typ"] = self.uri.object.typ

        ensure_dir(self._manifest_path.parent)
        ensure_file(
            self._manifest_path, yaml.safe_dump(_manifest, default_flow_style=False)
        )

    def add_fast_tag(self) -> None:
        _manifest = self._get_manifest()
        _seq = int(_manifest.get("fast_tag_seq", -1))
        _tag = _seq + 1
        while True:
            if f"v{_tag}" in _manifest["tags"]:
                _tag = _tag + 1
            else:
                break

        _manifest["fast_tag_seq"] = _tag
        self.add(tags=[f"v{_tag}", LATEST_TAG], manifest=_manifest)

    def add(
        self,
        tags: t.List[str],
        ignore_errors: bool = False,
        manifest: t.Optional[t.Dict] = None,
    ) -> None:
        _manifest = manifest or self._get_manifest()
        _version = self.uri.object.version

        if not _version and not ignore_errors:
            raise MissingFieldError(f"uri version, {self.uri}")

        for _t in tags:
            _t = _t.strip()
            if not _t:
                continue

            _ok, _reason = validate_obj_name(_t)
            if not _ok:
                if ignore_errors:
                    continue
                else:
                    raise FormatError(f"{_t}, reason:{_reason}")

            _pre_version = _manifest["tags"].get(_t, "")
            if _pre_version and _pre_version != _version:
                _manifest["versions"][_pre_version].pop(_t, None)
                if not _manifest["versions"][_pre_version]:
                    _manifest["versions"].pop(_pre_version, None)

            _manifest["tags"][_t] = _version
            if _version in _manifest["versions"]:
                _manifest["versions"][_version][_t] = True
            else:
                _manifest["versions"][_version] = {_t: True}

        self._save_manifest(_manifest)

    def remove(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        _manifest = self._get_manifest()
        for _t in tags:
            _version = _manifest["tags"].pop(_t, "")

            if _version not in _manifest["versions"]:
                if ignore_errors:
                    continue
                else:
                    raise NotFoundError(f"tag:{_t}, version:{_version}")

            _manifest["versions"][_version].pop(_t, None)
            if not _manifest["versions"][_version]:
                _manifest["versions"].pop(_version, None)

        self._save_manifest(_manifest)

    def __iter__(self) -> t.Generator[str, None, None]:
        for tag in self.list():
            yield tag

    def list(self) -> t.List[str]:
        _manifest = self._get_manifest()
        _version = self.uri.object.version

        if _version:
            _tags = _manifest["versions"].get(_version, {}).keys()
        else:
            _tags = _manifest["tags"].keys()
        return list(_tags)

    @classmethod
    def get_manifest_by_dir(cls, dir: Path) -> t.Dict[str, t.Any]:
        _mf = dir / DEFAULT_MANIFEST_NAME
        if _mf.exists():
            return load_yaml(_mf) or {}
        else:
            return {}



import typing as t
import tempfile
from abc import ABCMeta, abstractmethod
from pathlib import Path

import yaml
from fs.tarfs import TarFS

from starwhale.utils import load_yaml
from starwhale.consts import (
    RECOVER_DIRNAME,
    SW_TMP_DIR_NAME,
    SHORT_VERSION_CNT,
    VERSION_PREFIX_CNT,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir, guess_real_path
from starwhale.utils.config import SWCliConfigMixed


class BundleField(t.NamedTuple):
    name: str = ""
    version: str = ""
    tags: t.List[str] = []
    path: Path = Path()
    is_removed: bool = False


class BaseStorage(metaclass=ABCMeta):
    def __init__(self, uri: URI) -> None:
        self.uri = uri
        self.sw_config = SWCliConfigMixed()
        self.project_dir = self.sw_config.rootdir / self.uri.project
        self.loc, self.id = self._guess()

        self.building = False
        self._tmp_dir: t.Union[None, Path] = None

    @abstractmethod
    def _guess(self) -> t.Tuple[Path, str]:
        raise NotImplementedError

    @property
    def recover_loc(self) -> Path:
        raise NotImplementedError

    @property
    def snapshot_workdir(self) -> Path:
        raise NotImplementedError

    @property
    def manifest_path(self) -> Path:
        raise NotImplementedError

    @property
    def bundle_type(self) -> str:
        raise NotImplementedError

    @property
    def uri_type(self) -> str:
        raise NotImplementedError

    @property
    def bundle_dir(self) -> Path:
        version = self.uri.object.version
        return (
            self.project_dir
            / self.uri_type
            / self.uri.object.name
            / version[:VERSION_PREFIX_CNT]
        )

    @property
    def bundle_path(self) -> Path:
        if self.uri.object.version:
            return self.bundle_dir / f"{self.uri.object.version}{self.bundle_type}"
        else:
            return self.bundle_dir

    @property
    def manifest(self) -> t.Dict[str, t.Any]:
        if not self.manifest_path.exists():
            return {}
        else:
            return load_yaml(self.manifest_path)  # type: ignore

    def _get_recover_snapshot_workdir_for_bundle(self) -> Path:
        version = self.uri.object.version
        return (
            self.project_dir
            / "workdir"
            / self.uri_type
            / RECOVER_DIRNAME
            / self.uri.object.name
            / version[:VERSION_PREFIX_CNT]
            / version
        )

    @property
    def tmp_dir(self) -> Path:
        if not self._tmp_dir:
            base = self.sw_config.rootdir / SW_TMP_DIR_NAME
            ensure_dir(base)
            self._tmp_dir = Path(tempfile.mkdtemp(dir=base))
        return self._tmp_dir

    def _get_snapshot_workdir_for_bundle(self) -> Path:
        if self.building:
            return self.tmp_dir
        version = self.uri.object.version
        return (
            self.project_dir
            / "workdir"
            / self.uri_type
            / self.uri.object.name
            / version[:VERSION_PREFIX_CNT]
            / version
        )

    def _get_recover_loc_for_bundle(self) -> Path:
        loc = self.project_dir / self.uri_type / RECOVER_DIRNAME / self.uri.object.name

        version = self.uri.object.version
        if version:
            loc = loc / version[:VERSION_PREFIX_CNT] / f"{version}{self.bundle_type}"

        return loc

    def _guess_for_bundle(self) -> t.Tuple[Path, str]:
        name = self.uri.object.name
        version = self.uri.object.version
        rootdir = self.project_dir / self.uri_type / name
        if version:
            _p, _v, _ok = guess_real_path(
                rootdir / version[:VERSION_PREFIX_CNT],
                version,
            )

            if not _ok:
                _manifest = StandaloneTag.get_manifest_by_dir(rootdir)
                _tag_version = _manifest.get("tags", {}).get(version, "")
                if _tag_version:
                    _p, _v, _ok = guess_real_path(
                        rootdir / _tag_version[:VERSION_PREFIX_CNT], _tag_version
                    )

            if _v.endswith(self.bundle_type):
                _v = _v.split(self.bundle_type)[0]

            self.uri.object.version = _v
            return _p, _v
        else:
            return self.project_dir / self.uri_type / name, name

    def iter_bundle_history(self) -> t.Generator[BundleField, None, None]:
        rootdir = self.project_dir / self.uri_type / self.uri.object.name
        _manifest = StandaloneTag.get_manifest_by_dir(rootdir)
        tags_map: t.Dict[str, t.Any] = _manifest.get("versions", {})

        for _path in rootdir.glob(f"**/*{self.bundle_type}"):
            if not _path.name.endswith(self.bundle_type):
                continue
            _rt_version = _path.name.split(self.bundle_type)[0]
            _tags = tags_map.get(_rt_version, {}).keys()
            yield BundleField(
                name=_path.name,
                version=_rt_version,
                tags=list(_tags),
                path=_path,
                is_removed=False,
            )

    @classmethod
    def iter_all_bundles(
        cls,
        project_uri: URI,
        bundle_type: str,
        uri_type: str,
    ) -> t.Generator[BundleField, None, None]:
        sw = SWCliConfigMixed()
        _obj_dir = sw.rootdir / project_uri.project / uri_type
        _tags_map = {}
        for _path in _obj_dir.glob(f"**/*{bundle_type}"):
            if not _path.name.endswith(bundle_type):
                continue

            _resource_name = _path.parent.parent.name
            if _resource_name not in _tags_map:
                _manifest = StandaloneTag.get_manifest_by_dir(_obj_dir / _resource_name)
                _tags_map[_resource_name] = _manifest.get("versions", {})

            _rt_version = _path.name.split(bundle_type)[0]
            _tags = _tags_map.get(_resource_name, {}).get(_rt_version, {}).keys()
            yield BundleField(
                name=_resource_name,
                version=_rt_version,
                tags=list(_tags),
                path=_path,
                is_removed=RECOVER_DIRNAME in _path.parts,
            )

    @classmethod
    def get_manifest_by_path(
        cls, fpath: Path, bundle_type: str, uri_type: str, direct: bool = False
    ) -> t.Any:
        if not direct and fpath.name.endswith(bundle_type):
            _model_dir = fpath.parent.parent
            _project_dir = _model_dir.parent.parent

            _mname = _model_dir.name
            _mversion = fpath.name.split(bundle_type)[0]

            _extracted_dir = (
                _project_dir
                / "workdir"
                / uri_type
                / _mname
                / _mversion[:SHORT_VERSION_CNT]
                / _mversion
            )
            _extracted_manifest = _extracted_dir / DEFAULT_MANIFEST_NAME

            if _extracted_manifest.exists():
                return load_yaml(_extracted_manifest)

        with TarFS(str(fpath)) as tar:
            with tar.open(DEFAULT_MANIFEST_NAME) as f:
                return yaml.safe_load(f)

    @property
    def recover_snapshot_workdir(self) -> Path:
        return self._get_recover_snapshot_workdir_for_bundle()



import os
import sys
import shutil
import typing as t
import platform
import subprocess
from pathlib import Path, PurePath, PosixPath

import conda_pack
import virtualenv
from loguru import logger

from starwhale.utils import console, is_linux, venv_pack, get_downloadable_sw_version
from starwhale.consts import (
    ENV_VENV,
    ENV_CONDA,
    PythonRunEnv,
    ENV_CONDA_PREFIX,
    SW_PYPI_PKG_NAME,
    SW_DEV_DUMMY_VERSION,
    WHEEL_FILE_EXTENSION,
    DEFAULT_CONDA_CHANNEL,
)
from starwhale.version import STARWHALE_VERSION
from starwhale.utils.fs import ensure_dir, ensure_file, extract_tar
from starwhale.utils.error import (
    FormatError,
    ExistedError,
    NotFoundError,
    NoSupportError,
    ParameterError,
    MissingFieldError,
    PythonEnvironmentError,
)
from starwhale.utils.process import check_call

SUPPORTED_PIP_REQ = ["requirements.txt", "pip-req.txt", "pip3-req.txt"]

_DUMMY_FIELD = -1

_ConfigsT = t.Optional[t.Dict[str, t.Dict[str, t.Union[str, t.List[str]]]]]
_PipConfigT = t.Optional[t.Dict[str, t.Union[str, t.List[str]]]]
_PipReqT = t.Union[str, Path, PosixPath]


class EnvTarType:
    CONDA = "conda_env.tar.gz"
    VENV = "venv_env.tar.gz"


class PythonVersionField(t.NamedTuple):
    major: int = _DUMMY_FIELD
    minor: int = _DUMMY_FIELD
    micro: int = _DUMMY_FIELD


def conda_install_req(
    req: t.Union[str, Path],
    env_name: str = "",
    prefix_path: t.Optional[Path] = None,
    enable_pre: bool = False,
    use_pip_install: bool = True,
    configs: _ConfigsT = None,
) -> None:
    if not req:
        return

    configs = configs or {}
    prefix_cmd = [get_conda_bin(), "run" if use_pip_install else "install"]

    if env_name:
        prefix_cmd += ["--name", env_name]
    elif prefix_path is not None:
        prefix_cmd += ["--prefix", str(prefix_path.absolute())]

    if use_pip_install:
        prefix_cmd += ["python3", "-m", "pip"]
        _do_pip_install_req(prefix_cmd, req, enable_pre, configs.get("pip"))  # type: ignore
    else:
        channels = configs.get("conda", {}).get("channels", [DEFAULT_CONDA_CHANNEL])
        for _c in channels:
            prefix_cmd += ["--channel", _c]

        prefix_cmd += ["--yes", "--override-channels"]
        check_call(prefix_cmd + [str(req)])


def _do_pip_install_req(
    # TODO: support multiple reqs
    prefix_cmd: t.List[t.Any],
    req: t.Union[str, Path],
    enable_pre: bool = False,
    pip_config: _PipConfigT = None,
) -> None:
    cmd = prefix_cmd + [
        "install",
        "--exists-action",
        "w",
    ]

    pip_config = pip_config or {}
    _env = os.environ

    list_to_str: t.Callable[[t.Union[str, list]], str] = (
        lambda x: " ".join(x) if isinstance(x, list) else x
    )

    # TODO: remove SW_PYPI_* envs

    _index_url = _env.get("SW_PYPI_INDEX_URL", "")
    _extra_index_urls = [
        _env.get("SW_PYPI_EXTRA_INDEX_URL", ""),
        list_to_str(pip_config.get("extra_index_url", [])),
    ]
    _hosts = [
        _env.get("SW_PYPI_TRUSTED_HOST", ""),
        list_to_str(pip_config.get("trusted_host", [])),
    ]

    config_index_url = list_to_str(pip_config.get("index_url", ""))
    if _index_url:
        _extra_index_urls.append(config_index_url)
    else:
        _index_url = config_index_url

    _s_index = _index_url.strip()
    _s_extra_index = " ".join([s for s in _extra_index_urls if s.strip()])
    _s_hosts = " ".join([s for s in _hosts if s.strip()])

    if _s_index:
        cmd += ["--index-url", _s_index]
    if _s_extra_index:
        cmd += ["--extra-index-url", _s_extra_index]
    if _s_hosts:
        cmd += ["--trusted-host", _s_hosts]

    if enable_pre:
        cmd += ["--pre"]

    if isinstance(req, PurePath):
        if not req.name.endswith(WHEEL_FILE_EXTENSION):
            cmd += ["-r"]
        cmd += [str(req.absolute())]  # type: ignore
    elif os.path.isfile(req):
        if not req.endswith(WHEEL_FILE_EXTENSION):
            cmd += ["-r"]
        cmd += [req]
    else:
        cmd += [req]

    check_call(cmd)


def venv_install_req(
    venvdir: t.Union[str, Path],
    req: _PipReqT,
    enable_pre: bool = False,
    pip_config: _PipConfigT = None,
) -> None:
    if not req:
        return

    venvdir = str(venvdir)
    req = str(req)
    prefix_cmd = [os.path.join(venvdir, "bin", "pip")]
    _do_pip_install_req(prefix_cmd, req, enable_pre, pip_config)


def venv_activate(venvdir: t.Union[str, Path]) -> None:
    _fpath = Path(venvdir) / "bin" / "activate"
    cmd = f"source {_fpath.absolute()}"
    check_call(cmd, shell=True, executable="/bin/bash")


def render_python_env_activate(
    mode: str,
    prefix_path: Path,
    workdir: Path,
    local_packaged_env: bool = False,
    verbose: bool = True,
) -> None:
    if mode not in (PythonRunEnv.CONDA, PythonRunEnv.VENV):
        raise NoSupportError(f"mode({mode}) render python env activate scripts")

    if local_packaged_env:
        # conda local mode(conda-pack) should be activated by the source command.
        venv_activate_render(prefix_path, workdir, relocate=mode == PythonRunEnv.VENV)
    else:
        if mode == PythonRunEnv.CONDA:
            conda_activate_render(prefix_path, workdir, verbose=verbose)
        else:
            venv_activate_render(prefix_path, workdir, relocate=False, verbose=verbose)


def parse_python_version(s: str) -> PythonVersionField:
    s = s.strip().lower()

    if not s:
        raise FormatError("python version empty")

    if s.startswith("python"):
        s = s.split("python", 1)[-1]

    _vt = s.split(".")
    _major, _minor, _micro = int(_vt[0]), _DUMMY_FIELD, _DUMMY_FIELD

    if len(_vt) >= 2:
        _minor = int(_vt[1])

    if len(_vt) >= 3:
        _micro = int(_vt[2])

    return PythonVersionField(major=_major, minor=_minor, micro=_micro)


def venv_setup(
    venvdir: t.Union[str, Path],
    python_version: str,
    prompt: str = "",
    clear: bool = False,
) -> None:
    # TODO: define starwhale virtualenv.py
    # TODO: use more elegant method to make venv portable
    args = [str(venvdir)]
    if prompt:
        args += ["--prompt", prompt]

    if python_version:
        _v = parse_python_version(python_version)
        args += ["--python", f"{_v.major}.{_v.minor}"]

    if clear:
        args += ["--clear"]

    session = virtualenv.cli_run(args)
    console.print(f":clap: create venv@{venvdir}, python:{session.interpreter.version}")  # type: ignore


def get_user_pip_config_by_pybin(py_bin: str) -> t.Dict[str, str]:
    output = subprocess.check_output([py_bin, "-m", "pip", "config", "list", "--user"])
    config: t.Dict[str, str] = {}
    for line in output.decode().strip().split("\n"):
        line = line.strip()
        if not line or "=" not in line:
            continue

        k, v = line.split("=", 1)
        config[k] = v
    return config


def pip_freeze_by_pybin(
    py_bin: str,
    lock_fpath: t.Union[str, Path],
    include_editable: bool = False,
    emit_options: bool = False,
    include_local_wheel: bool = False,
) -> None:
    lock_fpath = Path(lock_fpath)
    logger.info(f"{py_bin}: pip freeze...")

    content = [f"# Generated by Starwhale({STARWHALE_VERSION}) Runtime Lock"]

    if not emit_options:
        _pip_config = get_user_pip_config_by_pybin(py_bin)
        if _pip_config.get("global.index-url"):
            content.append(f"--index-url {_pip_config['global.index-url']}")

        if _pip_config.get("global.extra-index-url"):
            content.append(f"--extra-index-url {_pip_config['global.extra-index-url']}")

        if _pip_config.get("install.trusted-host"):
            content.append(f"--trusted-host {_pip_config['install.trusted-host']}")

    ensure_file(lock_fpath, "\n".join(content) + "\n")
    cmd = [py_bin, "-m", "pip", "freeze", "--require-virtualenv"]
    if not include_editable:
        cmd += ["--exclude-editable"]
    cmd += [">>", str(lock_fpath)]

    check_call(" ".join(cmd), shell=True)

    if not include_local_wheel:
        content = []
        for line in lock_fpath.read_text().splitlines():
            if line.endswith(".whl") and "@ file://" in line:
                continue
            content.append(line)
        ensure_file(lock_fpath, "\n".join(content))


def pip_freeze(
    py_env: str,
    lock_fpath: t.Union[str, Path],
    include_editable: bool = False,
    emit_options: bool = False,
    include_local_wheel: bool = False,
) -> None:
    # TODO: add cmd timeout and error log
    _py_bin = get_user_runtime_python_bin(py_env)
    pip_freeze_by_pybin(
        py_bin=_py_bin,
        lock_fpath=lock_fpath,
        include_editable=include_editable,
        emit_options=emit_options,
        include_local_wheel=include_local_wheel,
    )


def user_pip_install_pkg(py_env: str, pkg_name: str, enable_pre: bool = False) -> None:
    _py_bin = get_user_runtime_python_bin(py_env)
    cmd = [_py_bin, "-m", "pip", "install"]

    if enable_pre:
        cmd += ["--pre"]

    cmd += [pkg_name]
    check_call(cmd)


def check_python_interpreter_consistency(mode: str) -> t.Tuple[bool, str, str]:
    if mode == PythonRunEnv.CONDA:
        ep_base_prefix = os.environ.get(ENV_CONDA_PREFIX, "")
    elif mode == PythonRunEnv.VENV:
        ep_base_prefix = os.environ.get(ENV_VENV, "")
    else:
        ep_base_prefix = (
            os.environ.get(ENV_VENV)
            or os.environ.get(ENV_CONDA_PREFIX)
            or sys.base_prefix
        )
    logger.debug(
        f"current python interpreter base_prefix:{sys.base_prefix}, expected env base_prefix:{ep_base_prefix}"
    )
    _ok = ep_base_prefix == sys.base_prefix
    if not _ok:
        cur_version = get_python_version()
        user_version = get_user_python_version(mode)
        if not user_version.startswith(cur_version):
            logger.error(
                f"swcli use python:{cur_version}, but runtime venv/conda python:{user_version}"
            )
            raise PythonEnvironmentError(
                f"swcli({cur_version}), runtime({user_version})"
            )

    return _ok, sys.base_prefix, ep_base_prefix


def guess_current_py_env() -> str:
    if is_venv():
        return PythonRunEnv.VENV
    elif is_conda():
        return PythonRunEnv.CONDA
    else:
        return PythonRunEnv.SYSTEM


def get_user_python_sys_paths(py_env: str) -> t.List[str]:
    logger.debug(f"get env({py_env}) sys path")
    _py_bin = get_user_runtime_python_bin(py_env)
    logger.info(f"{_py_bin}: sys.path")
    output = subprocess.check_output(
        [
            _py_bin,
            "-c",
            "import sys; print(','.join(sys.path))",
        ]
    )
    return output.decode().strip().split(",")


def conda_setup(
    python_version: str,
    name: str = "",
    prefix: t.Union[str, Path] = "",
    quiet: bool = False,
) -> None:
    if not name and not prefix:
        raise ParameterError("conda setup must set name or prefix")

    cmd = [get_conda_bin(), "create", "--yes"]
    if name:
        cmd += ["--name", name]

    if prefix:
        cmd += ["--prefix", str(prefix)]

    if quiet:
        cmd += ["--quiet"]

    cmd += [f"python={trunc_python_version(python_version)}"]
    check_call(cmd)


def conda_export(
    lock_fpath: t.Union[str, Path], name: str = "", prefix: str = ""
) -> None:
    # TODO: add cmd timeout
    cmd = [
        get_conda_bin(),
        "env",
        "export",
    ]

    if name:
        cmd += ["--name", name]

    if prefix:
        cmd += ["--prefix", prefix]

    cmd += ["--file", str(lock_fpath)]
    check_call(cmd)


def conda_env_update(
    env_fpath: t.Union[str, Path], target_env: t.Union[str, Path]
) -> None:
    target_env = Path(target_env).resolve()
    cmd = [
        get_conda_bin(),
        "env",
        "update",
        "--file",
        str(env_fpath),
        "--prefix",
        str(target_env),
    ]
    check_call(cmd)


def conda_activate(env: t.Union[str, Path]) -> None:
    cmd = [get_conda_bin(), "activate", str(env)]
    check_call(cmd)


def conda_activate_render(env_dir: Path, workdir: Path, verbose: bool = True) -> None:
    sw_cntr_content = """
_conda_hook="$(/opt/miniconda3/bin/conda shell.bash hook)"
cat >> /dev/stdout << EOF
$_conda_hook
conda activate /opt/starwhale/swmp/dep/conda/env
EOF
"""

    host_content = f"""
echo 'conda activate {env_dir.absolute()}'
"""
    _render_sw_activate(sw_cntr_content, host_content, workdir, verbose)


def venv_activate_render(
    venvdir: t.Union[str, Path],
    workdir: Path,
    relocate: bool = False,
    verbose: bool = True,
) -> None:
    bin = f"{venvdir}/bin"
    host_content = f"""
echo 'source {venvdir}/bin/activate'
"""

    if relocate:
        # TODO: support relocatable editable python package
        sw_cntr_content = f"""
sed -i '1d' {bin}/starwhale {bin}/sw {bin}/swcli {bin}/pip* {bin}/virtualenv
sed -i '1i\#!{bin}/python3' {bin}/starwhale {bin}/sw {bin}/swcli {bin}/pip* {bin}/virtualenv

sed -i 's#^VIRTUAL_ENV=.*$#VIRTUAL_ENV={venvdir}#g' {bin}/activate
rm -rf {bin}/python3
ln -s /usr/bin/python3 {bin}/python3
echo 'source {bin}/activate'
"""
    else:
        sw_cntr_content = host_content

    _render_sw_activate(sw_cntr_content, host_content, workdir, verbose)


def _render_sw_activate(
    sw_cntr_content: str, host_content: str, workdir: Path, verbose: bool = True
) -> None:
    _sw_path = workdir / "activate.sw"
    _host_path = workdir / "activate.host"

    ensure_file(_sw_path, sw_cntr_content, mode=0o755)
    ensure_file(_host_path, host_content, mode=0o755)

    if verbose:
        console.print(
            f" :clap: {_sw_path.name} and {_host_path.name} is generated at {workdir}"
        )
        console.print(" :compass: run cmd:  ")
        console.print(f" \t Docker Container: [bold red] $(sh {_sw_path}) [/]")
        console.print(f" \t Host: [bold red] $(sh {_host_path}) [/]")


def get_conda_prefix_path(name: str = "") -> str:
    cmd = [get_conda_bin(), "run"]
    if name:
        cmd += ["--name", name]

    cmd += ["printenv", "CONDA_PREFIX"]
    output = subprocess.check_output(cmd)
    return output.decode().strip()


def get_conda_bin() -> str:
    # TODO: add process cache
    for _p in (
        "/opt/miniconda3/bin/conda",
        "/opt/anaconda3/bin/conda",
        os.path.expanduser("~/miniconda3/bin/conda"),
        os.path.expanduser("~/anaconda3/bin/conda"),
    ):
        if os.path.exists(_p):
            return _p
    else:
        return "conda"


def package_python_env(
    export_dir: t.Union[str, Path],
    mode: str,
    env_prefix_path: str = "",
    env_name: str = "",
    include_editable: bool = False,
) -> bool:
    export_dir = Path(export_dir)

    sys_name = platform.system()
    if not is_linux():
        logger.warning(
            f"[info:dep]{sys_name} will skip conda/venv to generate local all bundles env"
        )
        return False

    ensure_dir(export_dir)
    dest_tar_path = export_dir / (
        EnvTarType.CONDA if mode == PythonRunEnv.CONDA else EnvTarType.VENV
    )

    if mode == PythonRunEnv.CONDA:
        env_name = env_name or get_conda_env()
        if not env_name and not env_prefix_path:
            raise Exception(
                "conda package must use env_name/env_prefix_path or be in the conda activated environment"
            )

        console.print(
            f":package: try to package conda env_name:{env_name}, env_prefix_path:{env_prefix_path}..."
        )

        conda_pack.pack(
            name=env_name,
            prefix=env_prefix_path,
            force=True,
            output=str(dest_tar_path),
            ignore_editable_packages=not include_editable,
        )
    elif mode == PythonRunEnv.VENV:
        env_prefix_path = env_prefix_path or get_venv_env()

        if include_editable:
            raise NoSupportError("venv exports editable packages")

        if not env_prefix_path:
            raise Exception(
                "virtualenv package must use env_prefix_path or be in the virtualenv activated environment"
            )

        console.print(f":package: try to package virtualenv {env_prefix_path}...")
        venv_pack.pack(
            prefix=env_prefix_path,
            force=True,
            output=str(dest_tar_path),
        )
    else:
        raise NoSupportError(f"package python env in {mode} mode")

    console.print(f":beer_mug: pack @ [underline]{dest_tar_path}[/]")
    return True


def activate_python_env(mode: str, identity: str, interactive: bool) -> None:
    if mode == PythonRunEnv.VENV:
        cmd = f"source {identity}/bin/activate"
    elif mode == PythonRunEnv.CONDA:
        cmd = f"conda activate {identity}"
    else:
        raise NoSupportError(mode)

    if interactive:
        import shellingham

        try:
            _name, _bin = shellingham.detect_shell()
        except shellingham.ShellDetectionFailure:
            _name, _bin = "", ""

        if not _bin.startswith("/") or _name == _bin:
            _bin = shutil.which(_name) or _bin

        if _name == "zsh":
            # https://zsh.sourceforge.io/Intro/intro_3.html
            os.execl(
                _bin,
                _bin,
                "-c",
                f"""temp_dir={identity} && \
                echo ". $HOME/.zshrc && {cmd}" > $temp_dir/.zshrc && \
                ZDOTDIR=$temp_dir zsh -i""",
            )
        elif _name == "bash":
            # https://www.gnu.org/software/bash/manual/html_node/Bash-Startup-Files.html
            os.execl(
                _bin, _bin, "-c", f'bash --rcfile <(echo ". "$HOME/.bashrc" && {cmd}")'
            )
        elif _name == "fish":
            # https://fishshell.com/docs/current/language.html#configuration
            os.execl(_bin, _bin, "-C", cmd)

    # user executes the command manually
    console.print(":cake: run command in shell :cake:")
    console.print(f"\t[red][bold]{cmd}")


def create_python_env(
    mode: str,
    name: str,
    isolated_env_dir: Path,
    python_version: str,
    force: bool = False,
) -> None:
    if mode == PythonRunEnv.VENV:
        if isolated_env_dir.exists() and not force:
            raise ExistedError(str(isolated_env_dir))

        logger.info(f"create venv @ {isolated_env_dir}...")
        venv_setup(isolated_env_dir, python_version=python_version, prompt=name)
    elif mode == PythonRunEnv.CONDA:
        logger.info(
            f"create conda {name}:{isolated_env_dir}, use python {python_version}..."
        )
        conda_setup(python_version, prefix=isolated_env_dir)
    else:
        raise NoSupportError(mode)


def get_python_version() -> str:
    return f"{sys.version_info.major}.{sys.version_info.minor}"


def get_python_version_by_bin(py_bin: str) -> str:
    logger.info(f"{py_bin}: python version")
    output = subprocess.check_output(
        [
            py_bin,
            "-c",
            "import sys; _v=sys.version_info;print(f'{_v.major}.{_v.minor}.{_v.micro}')",
        ]
    )
    return output.decode().strip()


def get_conda_pybin(prefix: t.Union[str, Path] = "", name: str = "") -> str:
    if prefix:
        return str(Path(prefix) / "bin" / "python3")

    if name:
        output = subprocess.check_output(
            [get_conda_bin(), "run", "-n", name, "which", "python3"]
        )
        return output.decode().strip()

    raise MissingFieldError("set prefix or name parameter")


def get_user_python_version(py_env: str) -> str:
    _py_bin = get_user_runtime_python_bin(py_env)
    return get_python_version_by_bin(_py_bin)


def get_user_runtime_python_bin(py_env: str) -> str:
    _prefix = get_base_prefix(py_env)
    _py_bin = os.path.join(_prefix, "bin", "python3")
    if not os.path.exists(_py_bin):
        raise NotFoundError(_py_bin)

    return _py_bin


def check_valid_venv_prefix(prefix: t.Union[str, Path]) -> bool:
    return (Path(prefix) / "pyvenv.cfg").exists()


def check_valid_conda_prefix(prefix: t.Union[str, Path]) -> bool:
    return (Path(prefix) / "conda-meta").exists()


def guess_python_env_mode(prefix: t.Union[str, Path]) -> str:
    if check_valid_venv_prefix(prefix):
        return PythonRunEnv.VENV
    elif check_valid_conda_prefix(prefix):
        return PythonRunEnv.CONDA
    else:
        raise NoSupportError(f"guess python env from paht: {prefix}")


def get_base_prefix(py_env: str) -> str:
    if py_env == PythonRunEnv.VENV:
        _path = os.environ.get(ENV_VENV, "")
    elif py_env == PythonRunEnv.CONDA:
        _path = os.environ.get(ENV_CONDA_PREFIX, "")
    else:
        _path = sys.prefix

    if _path and os.path.exists(_path):
        return _path
    else:
        raise NotFoundError(f"mode:{py_env}, base_prefix:{_path}")


def is_venv() -> bool:
    # TODO: refactor for get external venv attr
    output = subprocess.check_output(
        [
            "python3",
            "-c",
            "import sys; print(sys.prefix != (getattr(sys, 'base_prefix', None) or (getattr(sys, 'real_prefix', None) or sys.prefix)))",
            # noqa: E501
        ],
    )
    return "True" in output.decode() or get_venv_env() != ""


def get_venv_env() -> str:
    return os.environ.get(ENV_VENV, "")


def is_conda() -> bool:
    return get_conda_env() != "" and get_conda_env_prefix() != ""


def get_python_run_env(mode: str = PythonRunEnv.AUTO) -> str:
    if mode == PythonRunEnv.VENV:
        if is_venv():
            return PythonRunEnv.VENV
        else:
            raise EnvironmentError(
                "expected venv mode, but cannot find venv environment"
            )
    elif mode == PythonRunEnv.CONDA:
        if is_conda():
            return PythonRunEnv.CONDA
        else:
            raise EnvironmentError("expected conda mode, but cannot find conda envs")
    elif mode == PythonRunEnv.AUTO:
        if is_conda() and is_venv():
            raise EnvironmentError("find venv and conda both activate")

        if is_conda():
            return PythonRunEnv.CONDA
        elif is_venv():
            return PythonRunEnv.VENV
        else:
            return PythonRunEnv.SYSTEM
    else:
        raise NoSupportError(f"python run env: {mode}")


def get_conda_env() -> str:
    return os.environ.get(ENV_CONDA, "")


def get_conda_env_prefix() -> str:
    return os.environ.get(ENV_CONDA_PREFIX, "")


def install_starwhale(
    prefix_path: Path,
    mode: str,
    version: str = "",
    force: bool = False,
    configs: _ConfigsT = None,
) -> None:
    if version == "" or version == SW_DEV_DUMMY_VERSION:
        version = get_downloadable_sw_version()

    req = SW_PYPI_PKG_NAME
    if version:
        req = f"{req}=={version}"

    _existed = check_user_python_pkg_exists(
        str(prefix_path / "bin" / "python3"), SW_PYPI_PKG_NAME
    )
    if _existed and not force:
        logger.info(f"{SW_PYPI_PKG_NAME} has already be installed at {prefix_path}")
        return

    configs = configs or {}
    if mode == PythonRunEnv.CONDA:
        conda_install_req(
            req=req, prefix_path=prefix_path, use_pip_install=True, configs=configs
        )
    elif mode == PythonRunEnv.VENV:
        venv_install_req(venvdir=prefix_path, req=req, pip_config=configs.get("pip"))
    else:
        raise NoSupportError(f"mode({mode}) install {SW_PYPI_PKG_NAME}")


def extract_conda_pkg(workdir: Path, isolated_env_dir: t.Optional[Path] = None) -> None:
    export_dir = workdir / "export"
    export_tar_fpath = export_dir / EnvTarType.CONDA
    conda_dir = isolated_env_dir or export_dir / "conda"
    extract_tar(export_tar_fpath, conda_dir, force=True)
    # TODO: conda local bundle restore wheel?


def extract_venv_pkg(workdir: Path, isolated_env_dir: t.Optional[Path] = None) -> None:
    export_dir = workdir / "export"
    venv_dir = isolated_env_dir or export_dir / "venv"
    export_tar_fpath = export_dir / EnvTarType.VENV
    extract_tar(export_tar_fpath, venv_dir, force=True)


def check_user_python_pkg_exists(py_bin: str, pkg_name: str) -> bool:
    cmd = [
        py_bin,
        "-c",
        f"import pkg_resources; pkg_resources.get_distribution('{pkg_name}')",
    ]
    try:
        check_call(cmd)
    except subprocess.CalledProcessError:
        return False
    else:
        return True


def trunc_python_version(python_version: str) -> str:
    _tp = python_version.strip().split(".")
    # TODO: support python full version format: {major}:{minor}:{micro}
    return ".".join(_tp[:2])



from typing import Any, Set, Dict, List


class _DAGData:
    def __init__(self) -> None:
        self.__graph: Dict[Any, Set[Any]] = {}
        self.__graph_reverse: Dict[Any, Set[Any]] = {}

    def vertices(self) -> Set[Any]:
        return set(self.__graph.keys())

    def add_vertex(self, vertex: Any) -> None:
        if vertex not in self.__graph:
            self.__graph[vertex] = set()
            self.__graph_reverse[vertex] = set()

    def add_edge(self, v_from: Any, v_to: Any) -> None:
        self.__graph[v_from].add(v_to)
        self.__graph_reverse[v_to].add(v_from)

    def remove_edge(self, v_from: Any, v_to: Any) -> None:
        self.__graph[v_from].remove(v_to)
        self.__graph_reverse[v_to].remove(v_from)

    def successors(self, vertex: Any) -> Set[Any]:
        return self.__graph[vertex]

    def predecessors(self, vertex: Any) -> Set[Any]:
        return self.__graph_reverse[vertex]


class DAG:
    def __init__(self) -> None:
        self.__data = _DAGData()

    def _validate_vertex(self, *vertices: Any) -> None:
        for vtx in vertices:
            if vtx not in self.__data.vertices():
                raise RuntimeError(f"Vertex '{vtx}' does not belong to DAG")

    def _has_path_to(self, v_from: Any, v_to: Any) -> bool:
        if v_from == v_to:
            return True
        for vtx in self.__data.successors(v_from):
            if self._has_path_to(vtx, v_to):
                return True
        return False

    def vertices(self) -> Set[Any]:
        return self.__data.vertices()

    def add_vertex(self, *vertices: Any) -> None:
        for vtx in vertices:
            self.__data.add_vertex(vtx)

    def add_edge(self, v_from: Any, *v_tos: Any) -> None:
        self._validate_vertex(v_from, *v_tos)

        for v_to in v_tos:
            if self._has_path_to(v_to, v_from):
                raise RuntimeError(
                    f"If this edge from '{v_from}' to '{v_to}' is added, it will cause the graph to cycle"
                )
            self.__data.add_edge(v_from, v_to)

    def remove_edge(self, v_from: Any, v_to: Any) -> None:
        self._validate_vertex(v_from, v_to)
        if v_to not in self.__data.successors(v_from):
            raise RuntimeError(f"Edge not found from '{v_from}' to '{v_to}'")

        self.__data.remove_edge(v_from, v_to)

    def vertex_size(self) -> int:
        return len(self.__data.vertices())

    def edge_size(self) -> int:
        size = 0
        for vtx in self.__data.vertices():
            size += self.out_degree(vtx)
        return size

    def successors(self, vertex: Any) -> Set[Any]:
        self._validate_vertex(vertex)
        return self.__data.successors(vertex)

    def predecessors(self, vertex: Any) -> Set[Any]:
        self._validate_vertex(vertex)
        return self.__data.predecessors(vertex)

    def in_degree(self, vertex: Any) -> int:
        return len(self.predecessors(vertex))

    def out_degree(self, vertex: Any) -> int:
        return len(self.successors(vertex))

    def all_starts(self) -> Set[Any]:
        return set(vtx for vtx in self.__data.vertices() if self.in_degree(vtx) == 0)

    def all_terminals(self) -> Set[Any]:
        return set(vtx for vtx in self.__data.vertices() if self.out_degree(vtx) == 0)


def generate_dag(_vertices: List[str], _edges: Dict[str, str]) -> DAG:
    _dag = DAG()
    _dag.add_vertex(*_vertices)
    for _from, _to in _edges.items():
        _dag.add_edge(_from, _to)
    return _dag
